from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ..arctrl_core.arc_types import (ArcAssay, ArcStudy, ArcInvestigation)
from ..arctrl_core.comment import Comment
from ..arctrl_core.Helper.identifier import create_missing_identifier
from ..arctrl_core.ontology_annotation import OntologyAnnotation
from ..arctrl_core.ontology_source_reference import OntologySourceReference
from ..arctrl_core.person import Person
from ..arctrl_core.publication import Publication
from ..arctrl_core.Table.composite_cell import CompositeCell
from ..fable_library.date import (today, to_string)
from ..fable_library.list import (choose, of_array, FSharpList, unzip, empty)
from ..fable_library.option import (map, default_arg)
from ..fable_library.seq import (map as map_1, concat)
from ..fable_library.seq2 import distinct_by
from ..fable_library.types import Array
from ..fable_library.util import (IEnumerable_1, string_hash)
from ..thoth_json_core.decode import (object, IRequiredGetter, string, IOptionalGetter, resize_array, IGetters, list_1 as list_1_1)
from ..thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1)
from .assay import (encoder as encoder_4, decoder as decoder_6, encoder_compressed as encoder_compressed_1, decoder_compressed as decoder_compressed_1)
from .comment import (encoder as encoder_6, decoder as decoder_8, ROCrate_encoder as ROCrate_encoder_5, ROCrate_decoder as ROCrate_decoder_5, ISAJson_encoder as ISAJson_encoder_5, ISAJson_decoder as ISAJson_decoder_5)
from .context.rocrate.isa_investigation_context import context_jsonvalue
from .context.rocrate.rocrate_context import (conforms_to_jsonvalue, context_jsonvalue as context_jsonvalue_1)
from .decode import Decode_objectNoAdditionalProperties
from .encode import (try_include, try_include_seq)
from .ontology_source_reference import (encoder as encoder_1, decoder as decoder_3, ROCrate_encoder as ROCrate_encoder_1, ROCrate_decoder as ROCrate_decoder_2, ISAJson_encoder as ISAJson_encoder_1, ISAJson_decoder as ISAJson_decoder_2)
from .person import (encoder as encoder_3, decoder as decoder_5, ROCrate_encoder as ROCrate_encoder_3, ROCrate_decoder as ROCrate_decoder_4, ISAJson_encoder as ISAJson_encoder_3, ISAJson_decoder as ISAJson_decoder_4)
from .publication import (encoder as encoder_2, decoder as decoder_4, ROCrate_encoder as ROCrate_encoder_2, ROCrate_decoder as ROCrate_decoder_3, ISAJson_encoder as ISAJson_encoder_2, ISAJson_decoder as ISAJson_decoder_3)
from .study import (encoder as encoder_5, decoder as decoder_7, encoder_compressed as encoder_compressed_2, decoder_compressed as decoder_compressed_2, ROCrate_encoder as ROCrate_encoder_4, ROCrate_decoder as ROCrate_decoder_1, ISAJson_encoder as ISAJson_encoder_4, ISAJson_decoder as ISAJson_decoder_1)

__A_ = TypeVar("__A_")

def encoder(inv: ArcInvestigation) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], inv: Any=inv) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow3277(__unit: None=None, inv: Any=inv) -> IEncodable:
        value: str = inv.Identifier
        class ObjectExpr3276(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr3276()

    def _arrow3279(value_1: str, inv: Any=inv) -> IEncodable:
        class ObjectExpr3278(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                return helpers_1.encode_string(value_1)

        return ObjectExpr3278()

    def _arrow3281(value_3: str, inv: Any=inv) -> IEncodable:
        class ObjectExpr3280(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                return helpers_2.encode_string(value_3)

        return ObjectExpr3280()

    def _arrow3283(value_5: str, inv: Any=inv) -> IEncodable:
        class ObjectExpr3282(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                return helpers_3.encode_string(value_5)

        return ObjectExpr3282()

    def _arrow3285(value_7: str, inv: Any=inv) -> IEncodable:
        class ObjectExpr3284(IEncodable):
            def Encode(self, helpers_4: IEncoderHelpers_1[Any]) -> Any:
                return helpers_4.encode_string(value_7)

        return ObjectExpr3284()

    def _arrow3286(osr: OntologySourceReference, inv: Any=inv) -> IEncodable:
        return encoder_1(osr)

    def _arrow3287(oa: Publication, inv: Any=inv) -> IEncodable:
        return encoder_2(oa)

    def _arrow3288(person: Person, inv: Any=inv) -> IEncodable:
        return encoder_3(person)

    def _arrow3289(assay: ArcAssay, inv: Any=inv) -> IEncodable:
        return encoder_4(assay)

    def _arrow3290(study: ArcStudy, inv: Any=inv) -> IEncodable:
        return encoder_5(study)

    def _arrow3292(value_9: str, inv: Any=inv) -> IEncodable:
        class ObjectExpr3291(IEncodable):
            def Encode(self, helpers_5: IEncoderHelpers_1[Any]) -> Any:
                return helpers_5.encode_string(value_9)

        return ObjectExpr3291()

    def _arrow3293(comment: Comment, inv: Any=inv) -> IEncodable:
        return encoder_6(comment)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([("Identifier", _arrow3277()), try_include("Title", _arrow3279, inv.Title), try_include("Description", _arrow3281, inv.Description), try_include("SubmissionDate", _arrow3283, inv.SubmissionDate), try_include("PublicReleaseDate", _arrow3285, inv.PublicReleaseDate), try_include_seq("OntologySourceReferences", _arrow3286, inv.OntologySourceReferences), try_include_seq("Publications", _arrow3287, inv.Publications), try_include_seq("Contacts", _arrow3288, inv.Contacts), try_include_seq("Assays", _arrow3289, inv.Assays), try_include_seq("Studies", _arrow3290, inv.Studies), try_include_seq("RegisteredStudyIdentifiers", _arrow3292, inv.RegisteredStudyIdentifiers), try_include_seq("Comments", _arrow3293, inv.Comments)]))
    class ObjectExpr3294(IEncodable):
        def Encode(self, helpers_6: IEncoderHelpers_1[Any], inv: Any=inv) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_6))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers_6.encode_object(arg)

    return ObjectExpr3294()


def _arrow3307(get: IGetters) -> ArcInvestigation:
    def _arrow3295(__unit: None=None) -> str:
        object_arg: IRequiredGetter = get.Required
        return object_arg.Field("Identifier", string)

    def _arrow3296(__unit: None=None) -> str | None:
        object_arg_1: IOptionalGetter = get.Optional
        return object_arg_1.Field("Title", string)

    def _arrow3297(__unit: None=None) -> str | None:
        object_arg_2: IOptionalGetter = get.Optional
        return object_arg_2.Field("Description", string)

    def _arrow3298(__unit: None=None) -> str | None:
        object_arg_3: IOptionalGetter = get.Optional
        return object_arg_3.Field("SubmissionDate", string)

    def _arrow3299(__unit: None=None) -> str | None:
        object_arg_4: IOptionalGetter = get.Optional
        return object_arg_4.Field("PublicReleaseDate", string)

    def _arrow3300(__unit: None=None) -> Array[OntologySourceReference] | None:
        arg_11: Decoder_1[Array[OntologySourceReference]] = resize_array(decoder_3)
        object_arg_5: IOptionalGetter = get.Optional
        return object_arg_5.Field("OntologySourceReferences", arg_11)

    def _arrow3301(__unit: None=None) -> Array[Publication] | None:
        arg_13: Decoder_1[Array[Publication]] = resize_array(decoder_4)
        object_arg_6: IOptionalGetter = get.Optional
        return object_arg_6.Field("Publications", arg_13)

    def _arrow3302(__unit: None=None) -> Array[Person] | None:
        arg_15: Decoder_1[Array[Person]] = resize_array(decoder_5)
        object_arg_7: IOptionalGetter = get.Optional
        return object_arg_7.Field("Contacts", arg_15)

    def _arrow3303(__unit: None=None) -> Array[ArcAssay] | None:
        arg_17: Decoder_1[Array[ArcAssay]] = resize_array(decoder_6)
        object_arg_8: IOptionalGetter = get.Optional
        return object_arg_8.Field("Assays", arg_17)

    def _arrow3304(__unit: None=None) -> Array[ArcStudy] | None:
        arg_19: Decoder_1[Array[ArcStudy]] = resize_array(decoder_7)
        object_arg_9: IOptionalGetter = get.Optional
        return object_arg_9.Field("Studies", arg_19)

    def _arrow3305(__unit: None=None) -> Array[str] | None:
        arg_21: Decoder_1[Array[str]] = resize_array(string)
        object_arg_10: IOptionalGetter = get.Optional
        return object_arg_10.Field("RegisteredStudyIdentifiers", arg_21)

    def _arrow3306(__unit: None=None) -> Array[Comment] | None:
        arg_23: Decoder_1[Array[Comment]] = resize_array(decoder_8)
        object_arg_11: IOptionalGetter = get.Optional
        return object_arg_11.Field("Comments", arg_23)

    return ArcInvestigation(_arrow3295(), _arrow3296(), _arrow3297(), _arrow3298(), _arrow3299(), _arrow3300(), _arrow3301(), _arrow3302(), _arrow3303(), _arrow3304(), None, None, _arrow3305(), _arrow3306())


decoder: Decoder_1[ArcInvestigation] = object(_arrow3307)

def encoder_compressed(string_table: Any, oa_table: Any, cell_table: Any, inv: ArcInvestigation) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, inv: Any=inv) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow3311(__unit: None=None, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, inv: Any=inv) -> IEncodable:
        value: str = inv.Identifier
        class ObjectExpr3310(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr3310()

    def _arrow3313(value_1: str, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, inv: Any=inv) -> IEncodable:
        class ObjectExpr3312(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                return helpers_1.encode_string(value_1)

        return ObjectExpr3312()

    def _arrow3315(value_3: str, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, inv: Any=inv) -> IEncodable:
        class ObjectExpr3314(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                return helpers_2.encode_string(value_3)

        return ObjectExpr3314()

    def _arrow3317(value_5: str, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, inv: Any=inv) -> IEncodable:
        class ObjectExpr3316(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                return helpers_3.encode_string(value_5)

        return ObjectExpr3316()

    def _arrow3319(value_7: str, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, inv: Any=inv) -> IEncodable:
        class ObjectExpr3318(IEncodable):
            def Encode(self, helpers_4: IEncoderHelpers_1[Any]) -> Any:
                return helpers_4.encode_string(value_7)

        return ObjectExpr3318()

    def _arrow3320(osr: OntologySourceReference, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, inv: Any=inv) -> IEncodable:
        return encoder_1(osr)

    def _arrow3321(oa: Publication, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, inv: Any=inv) -> IEncodable:
        return encoder_2(oa)

    def _arrow3322(person: Person, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, inv: Any=inv) -> IEncodable:
        return encoder_3(person)

    def _arrow3323(assay: ArcAssay, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, inv: Any=inv) -> IEncodable:
        return encoder_compressed_1(string_table, oa_table, cell_table, assay)

    def _arrow3324(study: ArcStudy, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, inv: Any=inv) -> IEncodable:
        return encoder_compressed_2(string_table, oa_table, cell_table, study)

    def _arrow3326(value_9: str, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, inv: Any=inv) -> IEncodable:
        class ObjectExpr3325(IEncodable):
            def Encode(self, helpers_5: IEncoderHelpers_1[Any]) -> Any:
                return helpers_5.encode_string(value_9)

        return ObjectExpr3325()

    def _arrow3327(comment: Comment, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, inv: Any=inv) -> IEncodable:
        return encoder_6(comment)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([("Identifier", _arrow3311()), try_include("Title", _arrow3313, inv.Title), try_include("Description", _arrow3315, inv.Description), try_include("SubmissionDate", _arrow3317, inv.SubmissionDate), try_include("PublicReleaseDate", _arrow3319, inv.PublicReleaseDate), try_include_seq("OntologySourceReferences", _arrow3320, inv.OntologySourceReferences), try_include_seq("Publications", _arrow3321, inv.Publications), try_include_seq("Contacts", _arrow3322, inv.Contacts), try_include_seq("Assays", _arrow3323, inv.Assays), try_include_seq("Studies", _arrow3324, inv.Studies), try_include_seq("RegisteredStudyIdentifiers", _arrow3326, inv.RegisteredStudyIdentifiers), try_include_seq("Comments", _arrow3327, inv.Comments)]))
    class ObjectExpr3328(IEncodable):
        def Encode(self, helpers_6: IEncoderHelpers_1[Any], string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, inv: Any=inv) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_6))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers_6.encode_object(arg)

    return ObjectExpr3328()


def decoder_compressed(string_table: Array[str], oa_table: Array[OntologyAnnotation], cell_table: Array[CompositeCell]) -> Decoder_1[ArcInvestigation]:
    def _arrow3341(get: IGetters, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table) -> ArcInvestigation:
        def _arrow3329(__unit: None=None) -> str:
            object_arg: IRequiredGetter = get.Required
            return object_arg.Field("Identifier", string)

        def _arrow3330(__unit: None=None) -> str | None:
            object_arg_1: IOptionalGetter = get.Optional
            return object_arg_1.Field("Title", string)

        def _arrow3331(__unit: None=None) -> str | None:
            object_arg_2: IOptionalGetter = get.Optional
            return object_arg_2.Field("Description", string)

        def _arrow3332(__unit: None=None) -> str | None:
            object_arg_3: IOptionalGetter = get.Optional
            return object_arg_3.Field("SubmissionDate", string)

        def _arrow3333(__unit: None=None) -> str | None:
            object_arg_4: IOptionalGetter = get.Optional
            return object_arg_4.Field("PublicReleaseDate", string)

        def _arrow3334(__unit: None=None) -> Array[OntologySourceReference] | None:
            arg_11: Decoder_1[Array[OntologySourceReference]] = resize_array(decoder_3)
            object_arg_5: IOptionalGetter = get.Optional
            return object_arg_5.Field("OntologySourceReferences", arg_11)

        def _arrow3335(__unit: None=None) -> Array[Publication] | None:
            arg_13: Decoder_1[Array[Publication]] = resize_array(decoder_4)
            object_arg_6: IOptionalGetter = get.Optional
            return object_arg_6.Field("Publications", arg_13)

        def _arrow3336(__unit: None=None) -> Array[Person] | None:
            arg_15: Decoder_1[Array[Person]] = resize_array(decoder_5)
            object_arg_7: IOptionalGetter = get.Optional
            return object_arg_7.Field("Contacts", arg_15)

        def _arrow3337(__unit: None=None) -> Array[ArcAssay] | None:
            arg_17: Decoder_1[Array[ArcAssay]] = resize_array(decoder_compressed_1(string_table, oa_table, cell_table))
            object_arg_8: IOptionalGetter = get.Optional
            return object_arg_8.Field("Assays", arg_17)

        def _arrow3338(__unit: None=None) -> Array[ArcStudy] | None:
            arg_19: Decoder_1[Array[ArcStudy]] = resize_array(decoder_compressed_2(string_table, oa_table, cell_table))
            object_arg_9: IOptionalGetter = get.Optional
            return object_arg_9.Field("Studies", arg_19)

        def _arrow3339(__unit: None=None) -> Array[str] | None:
            arg_21: Decoder_1[Array[str]] = resize_array(string)
            object_arg_10: IOptionalGetter = get.Optional
            return object_arg_10.Field("RegisteredStudyIdentifiers", arg_21)

        def _arrow3340(__unit: None=None) -> Array[Comment] | None:
            arg_23: Decoder_1[Array[Comment]] = resize_array(decoder_8)
            object_arg_11: IOptionalGetter = get.Optional
            return object_arg_11.Field("Comments", arg_23)

        return ArcInvestigation(_arrow3329(), _arrow3330(), _arrow3331(), _arrow3332(), _arrow3333(), _arrow3334(), _arrow3335(), _arrow3336(), _arrow3337(), _arrow3338(), None, None, _arrow3339(), _arrow3340())

    return object(_arrow3341)


def ROCrate_genID(i: ArcInvestigation) -> str:
    return "./"


def ROCrate_encoder(oa: ArcInvestigation) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], oa: Any=oa) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow3345(__unit: None=None, oa: Any=oa) -> IEncodable:
        value: str = ROCrate_genID(oa)
        class ObjectExpr3344(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr3344()

    class ObjectExpr3346(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any], oa: Any=oa) -> Any:
            return helpers_1.encode_string("Investigation")

    class ObjectExpr3347(IEncodable):
        def Encode(self, helpers_2: IEncoderHelpers_1[Any], oa: Any=oa) -> Any:
            return helpers_2.encode_string("Investigation")

    def _arrow3349(__unit: None=None, oa: Any=oa) -> IEncodable:
        value_3: str = oa.Identifier
        class ObjectExpr3348(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                return helpers_3.encode_string(value_3)

        return ObjectExpr3348()

    def _arrow3351(__unit: None=None, oa: Any=oa) -> IEncodable:
        value_4: str = ArcInvestigation.FileName()
        class ObjectExpr3350(IEncodable):
            def Encode(self, helpers_4: IEncoderHelpers_1[Any]) -> Any:
                return helpers_4.encode_string(value_4)

        return ObjectExpr3350()

    def _arrow3353(value_5: str, oa: Any=oa) -> IEncodable:
        class ObjectExpr3352(IEncodable):
            def Encode(self, helpers_5: IEncoderHelpers_1[Any]) -> Any:
                return helpers_5.encode_string(value_5)

        return ObjectExpr3352()

    def _arrow3355(value_7: str, oa: Any=oa) -> IEncodable:
        class ObjectExpr3354(IEncodable):
            def Encode(self, helpers_6: IEncoderHelpers_1[Any]) -> Any:
                return helpers_6.encode_string(value_7)

        return ObjectExpr3354()

    def _arrow3357(value_9: str, oa: Any=oa) -> IEncodable:
        class ObjectExpr3356(IEncodable):
            def Encode(self, helpers_7: IEncoderHelpers_1[Any]) -> Any:
                return helpers_7.encode_string(value_9)

        return ObjectExpr3356()

    def _arrow3360(__unit: None=None, oa: Any=oa) -> IEncodable:
        def _arrow3358(__unit: None=None) -> str:
            copy_of_struct: Any = today()
            return to_string(copy_of_struct, "yyyy-MM-dd")

        value_12: str = default_arg(oa.PublicReleaseDate, _arrow3358())
        class ObjectExpr3359(IEncodable):
            def Encode(self, helpers_8: IEncoderHelpers_1[Any]) -> Any:
                return helpers_8.encode_string(value_12)

        return ObjectExpr3359()

    def _arrow3361(osr: OntologySourceReference, oa: Any=oa) -> IEncodable:
        return ROCrate_encoder_1(osr)

    def _arrow3362(oa_1: Publication, oa: Any=oa) -> IEncodable:
        return ROCrate_encoder_2(oa_1)

    def _arrow3363(oa_2: Person, oa: Any=oa) -> IEncodable:
        return ROCrate_encoder_3(oa_2)

    def _arrow3364(s: ArcStudy, oa: Any=oa) -> IEncodable:
        return ROCrate_encoder_4(None, s)

    def _arrow3365(comment: Comment, oa: Any=oa) -> IEncodable:
        return ROCrate_encoder_5(comment)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([("@id", _arrow3345()), ("@type", ObjectExpr3346()), ("additionalType", ObjectExpr3347()), ("identifier", _arrow3349()), ("filename", _arrow3351()), try_include("title", _arrow3353, oa.Title), try_include("description", _arrow3355, oa.Description), try_include("submissionDate", _arrow3357, oa.SubmissionDate), ("publicReleaseDate", _arrow3360()), try_include_seq("ontologySourceReferences", _arrow3361, oa.OntologySourceReferences), try_include_seq("publications", _arrow3362, oa.Publications), try_include_seq("people", _arrow3363, oa.Contacts), try_include_seq("studies", _arrow3364, oa.Studies), try_include_seq("comments", _arrow3365, oa.Comments), ("@context", context_jsonvalue)]))
    class ObjectExpr3366(IEncodable):
        def Encode(self, helpers_9: IEncoderHelpers_1[Any], oa: Any=oa) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_9))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers_9.encode_object(arg)

    return ObjectExpr3366()


def _arrow3378(get: IGetters) -> ArcInvestigation:
    identifier: str
    match_value: str | None
    object_arg: IOptionalGetter = get.Optional
    match_value = object_arg.Field("identifier", string)
    identifier = create_missing_identifier() if (match_value is None) else match_value
    def _arrow3367(__unit: None=None) -> FSharpList[tuple[ArcStudy, FSharpList[ArcAssay]]] | None:
        arg_3: Decoder_1[FSharpList[tuple[ArcStudy, FSharpList[ArcAssay]]]] = list_1_1(ROCrate_decoder_1)
        object_arg_1: IOptionalGetter = get.Optional
        return object_arg_1.Field("studies", arg_3)

    pattern_input: tuple[FSharpList[ArcStudy], FSharpList[FSharpList[ArcAssay]]] = unzip(default_arg(_arrow3367(), empty()))
    studies_raw: FSharpList[ArcStudy] = pattern_input[0]
    def projection(a: ArcAssay) -> str:
        return a.Identifier

    class ObjectExpr3369:
        @property
        def Equals(self) -> Callable[[str, str], bool]:
            def _arrow3368(x: str, y: str) -> bool:
                return x == y

            return _arrow3368

        @property
        def GetHashCode(self) -> Callable[[str], int]:
            return string_hash

    assays: Array[ArcAssay] = list(distinct_by(projection, concat(pattern_input[1]), ObjectExpr3369()))
    studies: Array[ArcStudy] = list(studies_raw)
    def mapping(a_1: ArcStudy) -> str:
        return a_1.Identifier

    study_identifiers: Array[str] = list(map_1(mapping, studies_raw))
    def _arrow3370(__unit: None=None) -> str | None:
        object_arg_2: IOptionalGetter = get.Optional
        return object_arg_2.Field("title", string)

    def _arrow3371(__unit: None=None) -> str | None:
        object_arg_3: IOptionalGetter = get.Optional
        return object_arg_3.Field("description", string)

    def _arrow3372(__unit: None=None) -> str | None:
        object_arg_4: IOptionalGetter = get.Optional
        return object_arg_4.Field("submissionDate", string)

    def _arrow3373(__unit: None=None) -> str | None:
        object_arg_5: IOptionalGetter = get.Optional
        return object_arg_5.Field("publicReleaseDate", string)

    def _arrow3374(__unit: None=None) -> Array[OntologySourceReference] | None:
        arg_13: Decoder_1[Array[OntologySourceReference]] = resize_array(ROCrate_decoder_2)
        object_arg_6: IOptionalGetter = get.Optional
        return object_arg_6.Field("ontologySourceReferences", arg_13)

    def _arrow3375(__unit: None=None) -> Array[Publication] | None:
        arg_15: Decoder_1[Array[Publication]] = resize_array(ROCrate_decoder_3)
        object_arg_7: IOptionalGetter = get.Optional
        return object_arg_7.Field("publications", arg_15)

    def _arrow3376(__unit: None=None) -> Array[Person] | None:
        arg_17: Decoder_1[Array[Person]] = resize_array(ROCrate_decoder_4)
        object_arg_8: IOptionalGetter = get.Optional
        return object_arg_8.Field("people", arg_17)

    def _arrow3377(__unit: None=None) -> Array[Comment] | None:
        arg_19: Decoder_1[Array[Comment]] = resize_array(ROCrate_decoder_5)
        object_arg_9: IOptionalGetter = get.Optional
        return object_arg_9.Field("comments", arg_19)

    return ArcInvestigation(identifier, _arrow3370(), _arrow3371(), _arrow3372(), _arrow3373(), _arrow3374(), _arrow3375(), _arrow3376(), assays, studies, None, None, study_identifiers, _arrow3377())


ROCrate_decoder: Decoder_1[ArcInvestigation] = object(_arrow3378)

def ROCrate_encodeRoCrate(oa: ArcInvestigation) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], oa: Any=oa) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow3382(value: str, oa: Any=oa) -> IEncodable:
        class ObjectExpr3381(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr3381()

    def _arrow3384(value_2: str, oa: Any=oa) -> IEncodable:
        class ObjectExpr3383(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                return helpers_1.encode_string(value_2)

        return ObjectExpr3383()

    def _arrow3385(oa_1: ArcInvestigation, oa: Any=oa) -> IEncodable:
        return ROCrate_encoder(oa_1)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([try_include("@type", _arrow3382, "CreativeWork"), try_include("@id", _arrow3384, "ro-crate-metadata.json"), try_include("about", _arrow3385, oa), ("conformsTo", conforms_to_jsonvalue), ("@context", context_jsonvalue_1)]))
    class ObjectExpr3386(IEncodable):
        def Encode(self, helpers_2: IEncoderHelpers_1[Any], oa: Any=oa) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_2))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers_2.encode_object(arg)

    return ObjectExpr3386()


ISAJson_allowedFields: FSharpList[str] = of_array(["@id", "filename", "identifier", "title", "description", "submissionDate", "publicReleaseDate", "ontologySourceReferences", "publications", "people", "studies", "comments", "@type", "@context"])

def ISAJson_encoder(id_map: Any | None, inv: ArcInvestigation) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], id_map: Any=id_map, inv: Any=inv) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow3390(__unit: None=None, id_map: Any=id_map, inv: Any=inv) -> IEncodable:
        value: str = ROCrate_genID(inv)
        class ObjectExpr3389(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr3389()

    def _arrow3392(__unit: None=None, id_map: Any=id_map, inv: Any=inv) -> IEncodable:
        value_1: str = ArcInvestigation.FileName()
        class ObjectExpr3391(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                return helpers_1.encode_string(value_1)

        return ObjectExpr3391()

    def _arrow3394(__unit: None=None, id_map: Any=id_map, inv: Any=inv) -> IEncodable:
        value_2: str = inv.Identifier
        class ObjectExpr3393(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                return helpers_2.encode_string(value_2)

        return ObjectExpr3393()

    def _arrow3396(value_3: str, id_map: Any=id_map, inv: Any=inv) -> IEncodable:
        class ObjectExpr3395(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                return helpers_3.encode_string(value_3)

        return ObjectExpr3395()

    def _arrow3398(value_5: str, id_map: Any=id_map, inv: Any=inv) -> IEncodable:
        class ObjectExpr3397(IEncodable):
            def Encode(self, helpers_4: IEncoderHelpers_1[Any]) -> Any:
                return helpers_4.encode_string(value_5)

        return ObjectExpr3397()

    def _arrow3400(value_7: str, id_map: Any=id_map, inv: Any=inv) -> IEncodable:
        class ObjectExpr3399(IEncodable):
            def Encode(self, helpers_5: IEncoderHelpers_1[Any]) -> Any:
                return helpers_5.encode_string(value_7)

        return ObjectExpr3399()

    def _arrow3402(value_9: str, id_map: Any=id_map, inv: Any=inv) -> IEncodable:
        class ObjectExpr3401(IEncodable):
            def Encode(self, helpers_6: IEncoderHelpers_1[Any]) -> Any:
                return helpers_6.encode_string(value_9)

        return ObjectExpr3401()

    def _arrow3403(osr: OntologySourceReference, id_map: Any=id_map, inv: Any=inv) -> IEncodable:
        return ISAJson_encoder_1(id_map, osr)

    def _arrow3404(oa: Publication, id_map: Any=id_map, inv: Any=inv) -> IEncodable:
        return ISAJson_encoder_2(id_map, oa)

    def _arrow3405(person: Person, id_map: Any=id_map, inv: Any=inv) -> IEncodable:
        return ISAJson_encoder_3(id_map, person)

    def _arrow3406(s: ArcStudy, id_map: Any=id_map, inv: Any=inv) -> IEncodable:
        return ISAJson_encoder_4(id_map, None, s)

    def _arrow3407(comment: Comment, id_map: Any=id_map, inv: Any=inv) -> IEncodable:
        return ISAJson_encoder_5(id_map, comment)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([("@id", _arrow3390()), ("filename", _arrow3392()), ("identifier", _arrow3394()), try_include("title", _arrow3396, inv.Title), try_include("description", _arrow3398, inv.Description), try_include("submissionDate", _arrow3400, inv.SubmissionDate), try_include("publicReleaseDate", _arrow3402, inv.PublicReleaseDate), try_include_seq("ontologySourceReferences", _arrow3403, inv.OntologySourceReferences), try_include_seq("publications", _arrow3404, inv.Publications), try_include_seq("people", _arrow3405, inv.Contacts), try_include_seq("studies", _arrow3406, inv.Studies), try_include_seq("comments", _arrow3407, inv.Comments)]))
    class ObjectExpr3408(IEncodable):
        def Encode(self, helpers_7: IEncoderHelpers_1[Any], id_map: Any=id_map, inv: Any=inv) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_7))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers_7.encode_object(arg)

    return ObjectExpr3408()


def _arrow3420(get: IGetters) -> ArcInvestigation:
    identifer: str
    match_value: str | None
    object_arg: IOptionalGetter = get.Optional
    match_value = object_arg.Field("identifier", string)
    identifer = create_missing_identifier() if (match_value is None) else match_value
    def _arrow3409(__unit: None=None) -> FSharpList[tuple[ArcStudy, FSharpList[ArcAssay]]] | None:
        arg_3: Decoder_1[FSharpList[tuple[ArcStudy, FSharpList[ArcAssay]]]] = list_1_1(ISAJson_decoder_1)
        object_arg_1: IOptionalGetter = get.Optional
        return object_arg_1.Field("studies", arg_3)

    pattern_input: tuple[FSharpList[ArcStudy], FSharpList[FSharpList[ArcAssay]]] = unzip(default_arg(_arrow3409(), empty()))
    studies_raw: FSharpList[ArcStudy] = pattern_input[0]
    def projection(a: ArcAssay) -> str:
        return a.Identifier

    class ObjectExpr3411:
        @property
        def Equals(self) -> Callable[[str, str], bool]:
            def _arrow3410(x: str, y: str) -> bool:
                return x == y

            return _arrow3410

        @property
        def GetHashCode(self) -> Callable[[str], int]:
            return string_hash

    assays: Array[ArcAssay] = list(distinct_by(projection, concat(pattern_input[1]), ObjectExpr3411()))
    studies: Array[ArcStudy] = list(studies_raw)
    def mapping(a_1: ArcStudy) -> str:
        return a_1.Identifier

    study_identifiers: Array[str] = list(map_1(mapping, studies_raw))
    def _arrow3412(__unit: None=None) -> str | None:
        object_arg_2: IOptionalGetter = get.Optional
        return object_arg_2.Field("title", string)

    def _arrow3413(__unit: None=None) -> str | None:
        object_arg_3: IOptionalGetter = get.Optional
        return object_arg_3.Field("description", string)

    def _arrow3414(__unit: None=None) -> str | None:
        object_arg_4: IOptionalGetter = get.Optional
        return object_arg_4.Field("submissionDate", string)

    def _arrow3415(__unit: None=None) -> str | None:
        object_arg_5: IOptionalGetter = get.Optional
        return object_arg_5.Field("publicReleaseDate", string)

    def _arrow3416(__unit: None=None) -> Array[OntologySourceReference] | None:
        arg_13: Decoder_1[Array[OntologySourceReference]] = resize_array(ISAJson_decoder_2)
        object_arg_6: IOptionalGetter = get.Optional
        return object_arg_6.Field("ontologySourceReferences", arg_13)

    def _arrow3417(__unit: None=None) -> Array[Publication] | None:
        arg_15: Decoder_1[Array[Publication]] = resize_array(ISAJson_decoder_3)
        object_arg_7: IOptionalGetter = get.Optional
        return object_arg_7.Field("publications", arg_15)

    def _arrow3418(__unit: None=None) -> Array[Person] | None:
        arg_17: Decoder_1[Array[Person]] = resize_array(ISAJson_decoder_4)
        object_arg_8: IOptionalGetter = get.Optional
        return object_arg_8.Field("people", arg_17)

    def _arrow3419(__unit: None=None) -> Array[Comment] | None:
        arg_19: Decoder_1[Array[Comment]] = resize_array(ISAJson_decoder_5)
        object_arg_9: IOptionalGetter = get.Optional
        return object_arg_9.Field("comments", arg_19)

    return ArcInvestigation(identifer, _arrow3412(), _arrow3413(), _arrow3414(), _arrow3415(), _arrow3416(), _arrow3417(), _arrow3418(), assays, studies, None, None, study_identifiers, _arrow3419())


ISAJson_decoder: Decoder_1[ArcInvestigation] = Decode_objectNoAdditionalProperties(ISAJson_allowedFields, _arrow3420)

__all__ = ["encoder", "decoder", "encoder_compressed", "decoder_compressed", "ROCrate_genID", "ROCrate_encoder", "ROCrate_decoder", "ROCrate_encodeRoCrate", "ISAJson_allowedFields", "ISAJson_encoder", "ISAJson_decoder"]




from __future__ import annotations
from typing import Any
from ...arctrl_core.Helper.identifier import create_missing_identifier
from ...fable_library.list import (FSharpList, is_empty, reduce)
from ...fable_library.option import value
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.seq import (to_list, delay, append, singleton, empty)
from ...fable_library.string_ import (to_text, printf)
from ...fable_library.util import IEnumerable_1
from ..helper import clean
from ..ldcontext import LDContext
from ..ldobject import LDNode

def _expr1473() -> TypeInfo:
    return class_type("ARCtrl.ROCrate.LDPostalAddress", None, LDPostalAddress)


class LDPostalAddress:
    @staticmethod
    def schema_type() -> str:
        return "http://schema.org/PostalAddress"

    @staticmethod
    def address_country() -> str:
        return "http://schema.org/addressCountry"

    @staticmethod
    def postal_code() -> str:
        return "http://schema.org/postalCode"

    @staticmethod
    def street_address() -> str:
        return "http://schema.org/streetAddress"

    @staticmethod
    def try_get_address_country_as_string(s: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = s.TryGetPropertyAsSingleton(LDPostalAddress.address_country(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def get_address_country_as_string(s: LDNode, context: LDContext | None=None) -> str:
        match_value: Any | None = s.TryGetPropertyAsSingleton(LDPostalAddress.address_country(), context)
        if match_value is None:
            raise Exception(("Could not access property `addressCountry` of object with @id `" + s.Id) + "`")

        elif str(type(value(match_value))) == "<class \'str\'>":
            n: str = value(match_value)
            return n

        else: 
            raise Exception(("Value of property `addressCountry` of object with @id `" + s.Id) + "` should have been a string")


    @staticmethod
    def set_address_country_as_string(s: LDNode, n: str) -> None:
        s.SetProperty(LDPostalAddress.address_country(), n)

    @staticmethod
    def try_get_postal_code_as_string(s: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = s.TryGetPropertyAsSingleton(LDPostalAddress.postal_code(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def get_postal_code_as_string(s: LDNode, context: LDContext | None=None) -> str:
        match_value: Any | None = s.TryGetPropertyAsSingleton(LDPostalAddress.postal_code(), context)
        if match_value is None:
            raise Exception(("Could not access property `postalCode` of object with @id `" + s.Id) + "`")

        elif str(type(value(match_value))) == "<class \'str\'>":
            n: str = value(match_value)
            return n

        else: 
            raise Exception(("Value of property `postalCode` of object with @id `" + s.Id) + "` should have been a string")


    @staticmethod
    def set_postal_code_as_string(s: LDNode, n: str) -> None:
        s.SetProperty(LDPostalAddress.postal_code(), n)

    @staticmethod
    def try_get_street_address_as_string(s: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = s.TryGetPropertyAsSingleton(LDPostalAddress.street_address(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def get_street_address_as_string(s: LDNode, context: LDContext | None=None) -> str:
        match_value: Any | None = s.TryGetPropertyAsSingleton(LDPostalAddress.street_address(), context)
        if match_value is None:
            raise Exception(("Could not access property `streetAddress` of object with @id `" + s.Id) + "`")

        elif str(type(value(match_value))) == "<class \'str\'>":
            n: str = value(match_value)
            return n

        else: 
            raise Exception(("Value of property `streetAddress` of object with @id `" + s.Id) + "` should have been a string")


    @staticmethod
    def set_street_address_as_string(s: LDNode, n: str) -> None:
        s.SetProperty(LDPostalAddress.street_address(), n)

    @staticmethod
    def gen_id(address_country: str | None=None, postal_code: str | None=None, street_address: str | None=None) -> str:
        def _arrow1471(__unit: None=None) -> IEnumerable_1[str]:
            def _arrow1470(__unit: None=None) -> IEnumerable_1[str]:
                def _arrow1469(__unit: None=None) -> IEnumerable_1[str]:
                    return singleton("streetAddress") if (street_address is not None) else empty()

                return append(singleton("postalCode") if (postal_code is not None) else empty(), delay(_arrow1469))

            return append(singleton("addressCountry") if (address_country is not None) else empty(), delay(_arrow1470))

        items: FSharpList[str] = to_list(delay(_arrow1471))
        def _arrow1472(__unit: None=None) -> str:
            def reduction(acc: str, x: str) -> str:
                return ((("" + acc) + "_") + x) + ""

            arg: str = reduce(reduction, items)
            return to_text(printf("#%s"))(arg)

        return clean(create_missing_identifier() if is_empty(items) else _arrow1472())

    @staticmethod
    def validate(o: LDNode, context: LDContext | None=None) -> bool:
        return o.HasType(LDPostalAddress.schema_type(), context)

    @staticmethod
    def create(id: str | None=None, address_country: str | None=None, postal_code: str | None=None, street_address: str | None=None, context: LDContext | None=None) -> LDNode:
        s: LDNode = LDNode(LDPostalAddress.gen_id(address_country, postal_code, street_address) if (id is None) else id, [LDPostalAddress.schema_type()], None, context)
        s.SetOptionalProperty(LDPostalAddress.address_country(), address_country)
        s.SetOptionalProperty(LDPostalAddress.postal_code(), postal_code)
        s.SetOptionalProperty(LDPostalAddress.street_address(), street_address)
        return s


LDPostalAddress_reflection = _expr1473

__all__ = ["LDPostalAddress_reflection"]




from __future__ import annotations
from typing import Any
from ...arctrl_core.Helper.collections_ import Option_fromValueWithDefault
from ...arctrl_core.Table.composite_cell import CompositeCell
from ...fable_library.option import (map, bind, default_arg)
from ...fable_library.types import Array

def term_from_string_cells(tsr_col: int | None, tan_col: int | None, cell_values: Array[str]) -> CompositeCell:
    def _arrow3493(i: int, tsr_col: Any=tsr_col, tan_col: Any=tan_col, cell_values: Any=cell_values) -> str:
        return cell_values[i]

    tan: str | None = map(_arrow3493, tan_col)
    def _arrow3494(i_1: int, tsr_col: Any=tsr_col, tan_col: Any=tan_col, cell_values: Any=cell_values) -> str:
        return cell_values[i_1]

    tsr: str | None = map(_arrow3494, tsr_col)
    return CompositeCell.create_term_from_string(cell_values[0], tsr, tan)


def unitized_from_string_cells(unit_col: int, tsr_col: int | None, tan_col: int | None, cell_values: Array[str]) -> CompositeCell:
    unit: str = cell_values[unit_col]
    def _arrow3495(i: int, unit_col: Any=unit_col, tsr_col: Any=tsr_col, tan_col: Any=tan_col, cell_values: Any=cell_values) -> str:
        return cell_values[i]

    tan: str | None = map(_arrow3495, tan_col)
    def _arrow3496(i_1: int, unit_col: Any=unit_col, tsr_col: Any=tsr_col, tan_col: Any=tan_col, cell_values: Any=cell_values) -> str:
        return cell_values[i_1]

    tsr: str | None = map(_arrow3496, tsr_col)
    return CompositeCell.create_unitized_from_string(cell_values[0], unit, tsr, tan)


def free_text_from_string_cells(cell_values: Array[str]) -> CompositeCell:
    return CompositeCell.create_free_text(cell_values[0])


def data_from_string_cells(format: int | None, selector_format: int | None, cell_values: Array[str]) -> CompositeCell:
    def _arrow3497(i: int, format: Any=format, selector_format: Any=selector_format, cell_values: Any=cell_values) -> str | None:
        return Option_fromValueWithDefault("", cell_values[i])

    format_1: str | None = bind(_arrow3497, format)
    def _arrow3498(i_1: int, format: Any=format, selector_format: Any=selector_format, cell_values: Any=cell_values) -> str | None:
        return Option_fromValueWithDefault("", cell_values[i_1])

    selector_format_1: str | None = bind(_arrow3498, selector_format)
    return CompositeCell.create_data_from_string(cell_values[0], format_1, selector_format_1)


def to_string_cells(is_term: bool, has_unit: bool, cell: CompositeCell) -> Array[str]:
    if cell.tag == 0:
        if has_unit:
            return [cell.fields[0].NameText, "", default_arg(cell.fields[0].TermSourceREF, ""), cell.fields[0].TermAccessionOntobeeUrl]

        else: 
            return [cell.fields[0].NameText, default_arg(cell.fields[0].TermSourceREF, ""), cell.fields[0].TermAccessionOntobeeUrl]


    elif cell.tag == 2:
        return [cell.fields[0], cell.fields[1].NameText, default_arg(cell.fields[1].TermSourceREF, ""), cell.fields[1].TermAccessionOntobeeUrl]

    elif cell.tag == 3:
        format: str = default_arg(cell.fields[0].Format, "")
        selector_format: str = default_arg(cell.fields[0].SelectorFormat, "")
        return [default_arg(cell.fields[0].Name, ""), format, selector_format]

    elif has_unit:
        return [cell.fields[0], "", "", ""]

    elif is_term:
        return [cell.fields[0], "", ""]

    else: 
        return [cell.fields[0]]



__all__ = ["term_from_string_cells", "unitized_from_string_cells", "free_text_from_string_cells", "data_from_string_cells", "to_string_cells"]




from __future__ import annotations
from typing import Any
from ..arctrl_core.template import Template
from ..fable_library.async_builder import (singleton, Async)
from ..fable_library.option import default_arg
from ..fable_library.types import Array
from .JsonIO.Table.templates import Templates_fromJsonString
from .WebRequest.web_request import download_file

def get_templates(url: str | None=None) -> Async[Array[Template]]:
    url_1: str = default_arg(url, "https://github.com/nfdi4plants/Swate-templates/releases/download/latest/templates_v2.0.0.json")
    def _arrow4072(__unit: None=None, url: Any=url) -> Async[Array[Template]]:
        def _arrow4071(_arg: str) -> Async[Array[Template]]:
            map_result: Array[Template] = Templates_fromJsonString(_arg)
            return singleton.Return(map_result)

        return singleton.Bind(download_file(url_1), _arrow4071)

    return singleton.Delay(_arrow4072)


__all__ = ["get_templates"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...fable_library.list import (empty, FSharpList, map, append)
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.string_ import (to_text, printf)
from .types import (SheetEntity_1_some_2B595, SheetEntity_1, SheetElement, Message__MapText_11D407F6, Message, SheetEntity_1__get_Messages)

def _expr318() -> TypeInfo:
    return class_type("FsSpreadsheet.DSL.SheetBuilder", None, SheetBuilder)


class SheetBuilder:
    def __init__(self, name: str) -> None:
        self.name: str = name


SheetBuilder_reflection = _expr318

def SheetBuilder__ctor_Z721C83C5(name: str) -> SheetBuilder:
    return SheetBuilder(name)


def SheetBuilder_get_Empty(__unit: None=None) -> SheetEntity_1[FSharpList[SheetElement]]:
    return SheetEntity_1_some_2B595(empty())


def SheetBuilder__SignMessages_3F89FFFC(this: SheetBuilder, messages: FSharpList[Message]) -> FSharpList[Message]:
    def mapping(m: Message, this: Any=this, messages: Any=messages) -> Message:
        def _arrow319(__unit: None=None, m: Any=m) -> Callable[[str], str]:
            clo_1: Callable[[str], str] = to_text(printf("In Sheet %s: %s"))(this.name)
            return clo_1

        return Message__MapText_11D407F6(m, _arrow319())

    return map(mapping, messages)


def SheetBuilder__Run_5EFE5A35(this: SheetBuilder, children: SheetEntity_1[FSharpList[SheetElement]]) -> SheetEntity_1[tuple[str, FSharpList[SheetElement]]]:
    if children.tag == 1:
        return SheetEntity_1(1, children.fields[0])

    elif children.tag == 2:
        return SheetEntity_1(2, children.fields[0])

    else: 
        return SheetEntity_1(0, (this.name, children.fields[0]), children.fields[1])



def SheetBuilder__Combine_6037FAE0(this: SheetBuilder, wx1: SheetEntity_1[FSharpList[SheetElement]], wx2: SheetEntity_1[FSharpList[SheetElement]]) -> SheetEntity_1[FSharpList[SheetElement]]:
    (pattern_matching_result, l1, l2, messages1, messages2, messages2_1, messages1_1, f1, messages1_2, messages2_2, f2, messages1_3, messages2_3, messages1_4, messages2_4) = (None, None, None, None, None, None, None, None, None, None, None, None, None, None, None)
    if wx1.tag == 2:
        if wx2.tag == 2:
            pattern_matching_result = 1
            messages2_1 = wx2.fields[0]

        else: 
            pattern_matching_result = 2
            messages1_1 = wx1.fields[0]


    elif wx1.tag == 1:
        if wx2.tag == 0:
            pattern_matching_result = 4
            f2 = wx2.fields[0]
            messages1_3 = wx1.fields[0]
            messages2_3 = wx2.fields[1]

        elif wx2.tag == 1:
            pattern_matching_result = 5
            messages1_4 = wx1.fields[0]
            messages2_4 = wx2.fields[0]

        else: 
            pattern_matching_result = 1
            messages2_1 = wx2.fields[0]


    elif wx2.tag == 2:
        pattern_matching_result = 1
        messages2_1 = wx2.fields[0]

    elif wx2.tag == 1:
        pattern_matching_result = 3
        f1 = wx1.fields[0]
        messages1_2 = wx1.fields[1]
        messages2_2 = wx2.fields[0]

    else: 
        pattern_matching_result = 0
        l1 = wx1.fields[0]
        l2 = wx2.fields[0]
        messages1 = wx1.fields[1]
        messages2 = wx2.fields[1]

    if pattern_matching_result == 0:
        return SheetEntity_1(0, append(l1, l2), append(messages1, messages2))

    elif pattern_matching_result == 1:
        return SheetEntity_1(2, append(SheetEntity_1__get_Messages(wx1), messages2_1))

    elif pattern_matching_result == 2:
        return SheetEntity_1(2, append(messages1_1, SheetEntity_1__get_Messages(wx2)))

    elif pattern_matching_result == 3:
        return SheetEntity_1(0, f1, append(messages1_2, messages2_2))

    elif pattern_matching_result == 4:
        return SheetEntity_1(0, f2, append(messages1_3, messages2_3))

    elif pattern_matching_result == 5:
        return SheetEntity_1(1, append(messages1_4, messages2_4))



__all__ = ["SheetBuilder_reflection", "SheetBuilder_get_Empty", "SheetBuilder__SignMessages_3F89FFFC", "SheetBuilder__Run_5EFE5A35", "SheetBuilder__Combine_6037FAE0"]




from __future__ import annotations
from openpyxl import Workbook
from typing import Any
from ..fable_library.array_ import iterate as iterate_1
from ..fable_library.seq import iterate
from ..fable_library.util import ignore
from ..fable_openpyxl.openpyxl import (Workbook as Workbook_1, Worksheet)
from ..fs_spreadsheet.fs_workbook import FsWorkbook
from ..fs_spreadsheet.fs_worksheet import FsWorksheet
from .worksheet import (from_fs_worksheet, to_fs_worksheet)

def from_fs_workbook(fs_wb: FsWorkbook) -> Workbook:
    FsWorkbook.validate_for_write(fs_wb)
    if len(fs_wb.GetWorksheets()) == 0:
        raise Exception("Workbook must contain at least one worksheet")

    py_wb: Workbook = Workbook(None)
    py_wb.remove(py_wb.active)
    def action(ws: FsWorksheet, fs_wb: Any=fs_wb) -> None:
        ignore(from_fs_worksheet(py_wb, ws))

    iterate(action, fs_wb.GetWorksheets())
    return py_wb


def to_fs_workbook(py_wb: Workbook) -> FsWorkbook:
    fs_wb: FsWorkbook = FsWorkbook()
    def action(ws: Worksheet, py_wb: Any=py_wb) -> None:
        if (len([list(inner_tuple) for inner_tuple in ws.values]) != 0) if (ws.title != "Sheet") else False:
            w: FsWorksheet = to_fs_worksheet(ws)
            w.RescanRows()
            value: None = fs_wb.AddWorksheet(w)
            ignore(None)


    iterate_1(action, py_wb.worksheets)
    return fs_wb


__all__ = ["from_fs_workbook", "to_fs_workbook"]




from __future__ import annotations
from typing import Any
from ...arctrl_core.Helper.identifier import create_missing_identifier
from ...fable_library.option import value
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.types import Array
from ..helper import clean
from ..ldcontext import LDContext
from ..ldobject import (LDNode, LDGraph)

def _expr1352() -> TypeInfo:
    return class_type("ARCtrl.ROCrate.LDFormalParameter", None, LDFormalParameter)


class LDFormalParameter:
    @staticmethod
    def schema_type() -> str:
        return "https://bioschemas.org/FormalParameter"

    @staticmethod
    def encoding_format() -> str:
        return "http://schema.org/encodingFormat"

    @staticmethod
    def name() -> str:
        return "http://schema.org/name"

    @staticmethod
    def same_as() -> str:
        return "http://schema.org/sameAs"

    @staticmethod
    def description() -> str:
        return "http://schema.org/description"

    @staticmethod
    def work_example() -> str:
        return "http://schema.org/workExample"

    @staticmethod
    def default_value() -> str:
        return "http://schema.org/defaultValue"

    @staticmethod
    def value_required() -> str:
        return "http://schema.org/valueRequired"

    @staticmethod
    def identifier() -> str:
        return "http://schema.org/identifier"

    @staticmethod
    def image() -> str:
        return "http://schema.org/image"

    @staticmethod
    def main_entity_of_page() -> str:
        return "http://schema.org/mainEntityOfPage"

    @staticmethod
    def potential_action() -> str:
        return "http://schema.org/potentialAction"

    @staticmethod
    def subject_of() -> str:
        return "http://schema.org/subjectOf"

    @staticmethod
    def url() -> str:
        return "http://schema.org/url"

    @staticmethod
    def alternate_name() -> str:
        return "http://schema.org/alternateName"

    @staticmethod
    def disambiguating_description() -> str:
        return "http://schema.org/disambiguatingDescription"

    @staticmethod
    def get_encoding_formats(fp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        return fp.GetPropertyNodes(LDFormalParameter.encoding_format(), None, graph, context)

    @staticmethod
    def set_encoding_formats(fp: LDNode, encoding_formats: Array[LDNode], context: LDContext | None=None) -> Any:
        return fp.SetProperty(LDFormalParameter.encoding_format(), encoding_formats, context)

    @staticmethod
    def try_get_name_as_string(fp: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = fp.TryGetPropertyAsSingleton(LDFormalParameter.name(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def get_name_as_string(fp: LDNode, context: LDContext | None=None) -> str:
        match_value: Any | None = fp.TryGetPropertyAsSingleton(LDFormalParameter.name(), context)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                n: str = value(match_value)
                return n

            else: 
                raise Exception(("Property of `name` of object with @id `" + fp.Id) + "` was not a string")


        else: 
            raise Exception(("Could not access property `name` of object with @id `" + fp.Id) + "`")


    @staticmethod
    def set_name_as_string(fp: LDNode, name: str, context: LDContext | None=None) -> Any:
        return fp.SetProperty(LDFormalParameter.name(), name, context)

    @staticmethod
    def get_same_as(fp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        return fp.GetPropertyNodes(LDFormalParameter.same_as(), None, graph, context)

    @staticmethod
    def set_same_as(fp: LDNode, sames: Array[LDNode], context: LDContext | None=None) -> Any:
        return fp.SetProperty(LDFormalParameter.same_as(), sames, context)

    @staticmethod
    def try_get_description_as_string(fp: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = fp.TryGetPropertyAsSingleton(LDFormalParameter.description(), context)
        (pattern_matching_result, d) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                d = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return d

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_description_as_string(fp: LDNode, description: str, context: LDContext | None=None) -> Any:
        return fp.SetProperty(LDFormalParameter.description(), description, context)

    @staticmethod
    def try_get_work_example_as_string(fp: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = fp.TryGetPropertyAsSingleton(LDFormalParameter.work_example(), context)
        (pattern_matching_result, we) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                we = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return we

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_work_example_as_string(fp: LDNode, work_example: str, context: LDContext | None=None) -> Any:
        return fp.SetProperty(LDFormalParameter.work_example(), work_example, context)

    @staticmethod
    def try_get_default_value_as_string(fp: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = fp.TryGetPropertyAsSingleton(LDFormalParameter.default_value(), context)
        (pattern_matching_result, dv) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                dv = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return dv

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_default_value_as_string(fp: LDNode, default_value: str, context: LDContext | None=None) -> Any:
        return fp.SetProperty(LDFormalParameter.default_value(), default_value, context)

    @staticmethod
    def try_get_value_required_as_boolean(fp: LDNode, context: LDContext | None=None) -> bool | None:
        match_value: Any | None = fp.TryGetPropertyAsSingleton(LDFormalParameter.value_required(), context)
        (pattern_matching_result, vr) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'bool\'>":
                pattern_matching_result = 0
                vr = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return vr

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_value_required_as_boolean(fp: LDNode, value_required: bool, context: LDContext | None=None) -> Any:
        return fp.SetProperty(LDFormalParameter.value_required(), value_required, context)

    @staticmethod
    def get_identifiers(fp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        return fp.GetPropertyNodes(LDFormalParameter.identifier(), None, graph, context)

    @staticmethod
    def set_identifiers(fp: LDNode, identifiers: Array[LDNode], context: LDContext | None=None) -> Any:
        return fp.SetProperty(LDFormalParameter.identifier(), identifiers, context)

    @staticmethod
    def try_get_image_as_string(fp: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = fp.TryGetPropertyAsSingleton(LDFormalParameter.image(), context)
        (pattern_matching_result, img) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                img = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return img

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_image_as_string(fp: LDNode, image: str, context: LDContext | None=None) -> Any:
        return fp.SetProperty(LDFormalParameter.image(), image, context)

    @staticmethod
    def try_get_main_entity_of_page_as_string(fp: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = fp.TryGetPropertyAsSingleton(LDFormalParameter.main_entity_of_page(), context)
        (pattern_matching_result, meop) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                meop = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return meop

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_main_entity_of_page_as_string(fp: LDNode, main_entity_of_page: str, context: LDContext | None=None) -> Any:
        return fp.SetProperty(LDFormalParameter.main_entity_of_page(), main_entity_of_page, context)

    @staticmethod
    def get_potential_actions(fp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        return fp.GetPropertyNodes(LDFormalParameter.potential_action(), None, graph, context)

    @staticmethod
    def set_potential_actions(fp: LDNode, potential_actions: Array[LDNode], context: LDContext | None=None) -> Any:
        return fp.SetProperty(LDFormalParameter.potential_action(), potential_actions, context)

    @staticmethod
    def get_subject_ofs(fp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        return fp.GetPropertyNodes(LDFormalParameter.subject_of(), None, graph, context)

    @staticmethod
    def set_subject_ofs(fp: LDNode, subject_ofs: Array[LDNode], context: LDContext | None=None) -> Any:
        return fp.SetProperty(LDFormalParameter.subject_of(), subject_ofs, context)

    @staticmethod
    def try_get_url_as_string(fp: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = fp.TryGetPropertyAsSingleton(LDFormalParameter.url(), context)
        (pattern_matching_result, url) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                url = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return url

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_url_as_string(fp: LDNode, url: str, context: LDContext | None=None) -> Any:
        return fp.SetProperty(LDFormalParameter.url(), url, context)

    @staticmethod
    def get_alternate_names(fp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        return fp.GetPropertyNodes(LDFormalParameter.alternate_name(), None, graph, context)

    @staticmethod
    def set_alternate_names(fp: LDNode, alternate_names: Array[LDNode], context: LDContext | None=None) -> Any:
        return fp.SetProperty(LDFormalParameter.alternate_name(), alternate_names, context)

    @staticmethod
    def try_get_disambiguating_description_as_string(fp: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = fp.TryGetPropertyAsSingleton(LDFormalParameter.disambiguating_description(), context)
        (pattern_matching_result, desc) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                desc = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return desc

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_disambiguating_description_as_string(fp: LDNode, disambiguating_description: str, context: LDContext | None=None) -> Any:
        return fp.SetProperty(LDFormalParameter.disambiguating_description(), disambiguating_description, context)

    @staticmethod
    def gen_id(name: str, workflow_name: str | None=None, run_name: str | None=None) -> str:
        def _arrow1347(__unit: None=None) -> str:
            run_1: str = run_name
            return ((("#FormalParameter_R_" + run_1) + "_") + name) + ""

        def _arrow1348(__unit: None=None) -> str:
            workflow_1: str = workflow_name
            return ((("#FormalParameter_W_" + workflow_1) + "_") + name) + ""

        def _arrow1349(__unit: None=None) -> str:
            run: str = run_name
            workflow: str = workflow_name
            return ((((("#FormalParameter_W_" + workflow) + "_R_") + run) + "_") + name) + ""

        return clean((_arrow1347() if (run_name is not None) else (((("#FormalParameter_" + name) + "_") + create_missing_identifier()) + "")) if (workflow_name is None) else (_arrow1348() if (run_name is None) else _arrow1349()))

    @staticmethod
    def validate(fp: LDNode, context: LDContext | None=None) -> bool:
        return (len(fp.AdditionalType) > 0) if fp.HasType(LDFormalParameter.schema_type(), context) else False

    @staticmethod
    def create(additional_type: str, id: str | None=None, encoding_formats: Array[LDNode] | None=None, name: str | None=None, same_as: Array[LDNode] | None=None, description: str | None=None, work_example: str | None=None, default_value: str | None=None, value_required: bool | None=None, identifiers: Array[LDNode] | None=None, image: str | None=None, subject_ofs: Array[LDNode] | None=None, url: str | None=None, alternate_names: Array[LDNode] | None=None, disambiguating_description: str | None=None, context: LDContext | None=None) -> LDNode:
        fp: LDNode = LDNode(clean(("#FormalParameter_" + create_missing_identifier()) + "") if (id is None) else id, [LDFormalParameter.schema_type()], None, context)
        fp.AdditionalType = [additional_type]
        fp.SetOptionalProperty(LDFormalParameter.encoding_format(), encoding_formats, context)
        fp.SetOptionalProperty(LDFormalParameter.name(), name, context)
        fp.SetOptionalProperty(LDFormalParameter.same_as(), same_as, context)
        fp.SetOptionalProperty(LDFormalParameter.description(), description, context)
        fp.SetOptionalProperty(LDFormalParameter.work_example(), work_example, context)
        fp.SetOptionalProperty(LDFormalParameter.default_value(), default_value, context)
        fp.SetOptionalProperty(LDFormalParameter.value_required(), value_required, context)
        fp.SetOptionalProperty(LDFormalParameter.identifier(), identifiers, context)
        fp.SetOptionalProperty(LDFormalParameter.image(), image, context)
        fp.SetOptionalProperty(LDFormalParameter.subject_of(), subject_ofs, context)
        fp.SetOptionalProperty(LDFormalParameter.url(), url, context)
        fp.SetOptionalProperty(LDFormalParameter.alternate_name(), alternate_names, context)
        fp.SetOptionalProperty(LDFormalParameter.disambiguating_description(), disambiguating_description, context)
        return fp


LDFormalParameter_reflection = _expr1352

__all__ = ["LDFormalParameter_reflection"]




from __future__ import annotations
from abc import abstractmethod
from typing import Protocol

class RequestBase(Protocol):
    @property
    @abstractmethod
    def url(self) -> str:
        ...


class Request(RequestBase):
    pass

class Flask(Protocol):
    pass

class FlaskStatic(Protocol):
    @abstractmethod
    def Create(self, name: str, static_url_path: str) -> Flask:
        ...


class IExports(Protocol):
    @property
    @abstractmethod
    def request(self) -> Request:
        ...

    @abstractmethod
    def url_for(self, route: str, filename: str) -> str:
        ...





from __future__ import annotations
from ..fable_library.option import default_arg
from ..fable_library.reflection import (TypeInfo, class_type)
from ..fable_library.result import FSharpResult_2
from ..fable_library.string_ import (to_fail, printf)
from ..fs_spreadsheet.fs_workbook import FsWorkbook
from ..fs_spreadsheet.Json.workbook import (decode_rows, encode_rows, decode_columns, encode_columns)
from ..thoth_json_python.decode import Decode_fromString
from ..thoth_json_python.encode import to_string

def _expr391() -> TypeInfo:
    return class_type("FsSpreadsheet.Py.Json", None, Json)


class Json:
    @staticmethod
    def try_from_rows_json_string(json: str) -> FSharpResult_2[FsWorkbook, str]:
        return Decode_fromString(decode_rows, json)

    @staticmethod
    def from_rows_json_string(json: str) -> FsWorkbook:
        match_value: FSharpResult_2[FsWorkbook, str] = Json.try_from_rows_json_string(json)
        return to_fail(printf("Could not deserialize json Workbook: \n%s"))(match_value.fields[0]) if (match_value.tag == 1) else match_value.fields[0]

    @staticmethod
    def to_rows_json_string(wb: FsWorkbook, spaces: int | None=None, no_numbering: bool | None=None) -> str:
        no_numbering_1: bool = default_arg(no_numbering, False)
        return to_string(default_arg(spaces, 2), encode_rows(no_numbering_1, wb))

    @staticmethod
    def try_from_columns_json_string(json: str) -> FSharpResult_2[FsWorkbook, str]:
        return Decode_fromString(decode_columns, json)

    @staticmethod
    def from_columns_json_string(json: str) -> FsWorkbook:
        match_value: FSharpResult_2[FsWorkbook, str] = Json.try_from_columns_json_string(json)
        return to_fail(printf("Could not deserialize json Workbook: \n%s"))(match_value.fields[0]) if (match_value.tag == 1) else match_value.fields[0]

    @staticmethod
    def to_columns_json_string(wb: FsWorkbook, spaces: int | None=None, no_numbering: bool | None=None) -> str:
        return to_string(default_arg(spaces, 2), encode_columns(default_arg(no_numbering, False), wb))


Json_reflection = _expr391

__all__ = ["Json_reflection"]




from __future__ import annotations
from dataclasses import dataclass
from typing import (Any, TypeVar)
from ....fable_library.reflection import (TypeInfo, string_type, record_type)
from ....fable_library.seq import map
from ....fable_library.types import Record
from ....fable_library.util import (to_enumerable, IEnumerable_1)
from ....thoth_json_core.types import (IEncodable, IEncoderHelpers_1)

__A_ = TypeVar("__A_")

def _expr1769() -> TypeInfo:
    return record_type("ARCtrl.Json.ROCrateContext.ProcessParameterValue.IContext", [], IContext, lambda: [("sdo", string_type), ("arc", string_type), ("ProcessParameterValue", string_type), ("ArcProcessParameterValue", string_type), ("category", string_type), ("value", string_type), ("unit", string_type)])


@dataclass(eq = False, repr = False, slots = True)
class IContext(Record):
    sdo: str
    arc: str
    ProcessParameterValue: str
    ArcProcessParameterValue: str
    category: str
    value: str
    unit: str

IContext_reflection = _expr1769

def _arrow1780(__unit: None=None) -> IEncodable:
    class ObjectExpr1770(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
            return helpers.encode_string("http://schema.org/")

    class ObjectExpr1771(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
            return helpers_1.encode_string("sdo:additionalType")

    class ObjectExpr1772(IEncodable):
        def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
            return helpers_2.encode_string("sdo:PropertyValue")

    class ObjectExpr1773(IEncodable):
        def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
            return helpers_3.encode_string("sdo:name")

    class ObjectExpr1774(IEncodable):
        def Encode(self, helpers_4: IEncoderHelpers_1[Any]) -> Any:
            return helpers_4.encode_string("sdo:propertyID")

    class ObjectExpr1775(IEncodable):
        def Encode(self, helpers_5: IEncoderHelpers_1[Any]) -> Any:
            return helpers_5.encode_string("sdo:value")

    class ObjectExpr1776(IEncodable):
        def Encode(self, helpers_6: IEncoderHelpers_1[Any]) -> Any:
            return helpers_6.encode_string("sdo:valueReference")

    class ObjectExpr1777(IEncodable):
        def Encode(self, helpers_7: IEncoderHelpers_1[Any]) -> Any:
            return helpers_7.encode_string("sdo:unitText")

    class ObjectExpr1778(IEncodable):
        def Encode(self, helpers_8: IEncoderHelpers_1[Any]) -> Any:
            return helpers_8.encode_string("sdo:unitCode")

    values: IEnumerable_1[tuple[str, IEncodable]] = to_enumerable([("sdo", ObjectExpr1770()), ("additionalType", ObjectExpr1771()), ("ProcessParameterValue", ObjectExpr1772()), ("category", ObjectExpr1773()), ("categoryCode", ObjectExpr1774()), ("value", ObjectExpr1775()), ("valueCode", ObjectExpr1776()), ("unit", ObjectExpr1777()), ("unitCode", ObjectExpr1778())])
    class ObjectExpr1779(IEncodable):
        def Encode(self, helpers_9: IEncoderHelpers_1[Any]) -> Any:
            def mapping(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers_9))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping, values)
            return helpers_9.encode_object(arg)

    return ObjectExpr1779()


context_jsonvalue: IEncodable = _arrow1780()

__all__ = ["IContext_reflection", "context_jsonvalue"]




from __future__ import annotations
from typing import (Any, TypeVar)
from ....fable_library.seq import map
from ....fable_library.util import (to_enumerable, IEnumerable_1)
from ....thoth_json_core.types import (IEncodable, IEncoderHelpers_1)

__A_ = TypeVar("__A_")

def _arrow1596(__unit: None=None) -> IEncodable:
    class ObjectExpr1583(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
            return helpers.encode_string("http://schema.org/")

    class ObjectExpr1584(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
            return helpers_1.encode_string("sdo:additionalType")

    class ObjectExpr1585(IEncodable):
        def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
            return helpers_2.encode_string("sdo:alternateName")

    class ObjectExpr1586(IEncodable):
        def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
            return helpers_3.encode_string("sdo:measurementMethod")

    class ObjectExpr1587(IEncodable):
        def Encode(self, helpers_4: IEncoderHelpers_1[Any]) -> Any:
            return helpers_4.encode_string("sdo:description")

    class ObjectExpr1588(IEncodable):
        def Encode(self, helpers_5: IEncoderHelpers_1[Any]) -> Any:
            return helpers_5.encode_string("sdo:name")

    class ObjectExpr1589(IEncodable):
        def Encode(self, helpers_6: IEncoderHelpers_1[Any]) -> Any:
            return helpers_6.encode_string("sdo:propertyID")

    class ObjectExpr1590(IEncodable):
        def Encode(self, helpers_7: IEncoderHelpers_1[Any]) -> Any:
            return helpers_7.encode_string("sdo:value")

    class ObjectExpr1591(IEncodable):
        def Encode(self, helpers_8: IEncoderHelpers_1[Any]) -> Any:
            return helpers_8.encode_string("sdo:valueReference")

    class ObjectExpr1592(IEncodable):
        def Encode(self, helpers_9: IEncoderHelpers_1[Any]) -> Any:
            return helpers_9.encode_string("sdo:unitText")

    class ObjectExpr1593(IEncodable):
        def Encode(self, helpers_10: IEncoderHelpers_1[Any]) -> Any:
            return helpers_10.encode_string("sdo:unitCode")

    class ObjectExpr1594(IEncodable):
        def Encode(self, helpers_11: IEncoderHelpers_1[Any]) -> Any:
            return helpers_11.encode_string("sdo:disambiguatingDescription")

    values: IEnumerable_1[tuple[str, IEncodable]] = to_enumerable([("sdo", ObjectExpr1583()), ("additionalType", ObjectExpr1584()), ("alternateName", ObjectExpr1585()), ("measurementMethod", ObjectExpr1586()), ("description", ObjectExpr1587()), ("category", ObjectExpr1588()), ("categoryCode", ObjectExpr1589()), ("value", ObjectExpr1590()), ("valueCode", ObjectExpr1591()), ("unit", ObjectExpr1592()), ("unitCode", ObjectExpr1593()), ("comments", ObjectExpr1594())])
    class ObjectExpr1595(IEncodable):
        def Encode(self, helpers_12: IEncoderHelpers_1[Any]) -> Any:
            def mapping(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers_12))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping, values)
            return helpers_12.encode_object(arg)

    return ObjectExpr1595()


context_jsonvalue: IEncodable = _arrow1596()

__all__ = ["context_jsonvalue"]




from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ..fable_library.array_ import append
from ..fable_library.async_ import (sequential, catch_async)
from ..fable_library.async_builder import (singleton, Async)
from ..fable_library.result import FSharpResult_2
from ..fable_library.types import (Array, to_string)
from ..fable_library.util import (IEnumerator, get_enumerator, IEnumerable_1)

_U = TypeVar("_U")

_T = TypeVar("_T")

__A = TypeVar("__A")

__B = TypeVar("__B")

def start_sequential(starter_f: Callable[[_T], Async[_U]], tasks: IEnumerable_1[Any]) -> Async[Array[Any]]:
    def loop(en: IEnumerator[_T], starter_f: Any=starter_f, tasks: Any=tasks) -> Async[Array[_U]]:
        def _arrow3771(__unit: None=None, en: Any=en) -> Async[Array[_U]]:
            def _arrow3770(_arg: _U | None=None) -> Async[Array[_U]]:
                def _arrow3769(_arg_1: Array[_U]) -> Async[Array[_U]]:
                    return singleton.Return(append([_arg], _arg_1, None))

                return singleton.Bind(loop(en), _arrow3769)

            return singleton.Bind(starter_f(en.System_Collections_Generic_IEnumerator_1_get_Current()), _arrow3770) if en.System_Collections_IEnumerator_MoveNext() else singleton.Return([])

        return singleton.Delay(_arrow3771)

    return loop(get_enumerator(tasks))


def all(tasks: IEnumerable_1[Async[Any]]) -> Async[Array[Any]]:
    return sequential(tasks)


def map(f: Callable[[__A], __B], v: Async[Any]) -> Async[Any]:
    def _arrow3773(__unit: None=None, f: Any=f, v: Any=v) -> Async[__B]:
        def _arrow3772(_arg: __A | None=None) -> Async[__B]:
            return singleton.Return(f(_arg))

        return singleton.Bind(v, _arrow3772)

    return singleton.Delay(_arrow3773)


def as_async(v: Async[Any]) -> Async[Any]:
    return v


def catch_with(f: Callable[[Exception], _T], p: Async[Any]) -> Async[Any]:
    def _arrow3775(__unit: None=None, f: Any=f, p: Any=p) -> Async[_T]:
        def _arrow3774(_arg: Any) -> Async[_T]:
            r: Any = _arg
            return singleton.Return(f(r.fields[0])) if (r.tag == 1) else singleton.Return(r.fields[0])

        return singleton.Bind(catch_async(p), _arrow3774)

    return singleton.Delay(_arrow3775)


def catch_as_result(p: Async[Any]) -> Async[FSharpResult_2[Any, str]]:
    def _arrow3777(__unit: None=None, p: Any=p) -> Async[FSharpResult_2[_T, str]]:
        def _arrow3776(_arg: Any) -> Async[FSharpResult_2[_T, str]]:
            r: Any = _arg
            return singleton.Return(FSharpResult_2(1, to_string(r.fields[0]))) if (r.tag == 1) else singleton.Return(FSharpResult_2(0, r.fields[0]))

        return singleton.Bind(catch_async(p), _arrow3776)

    return singleton.Delay(_arrow3777)


__all__ = ["start_sequential", "all", "map", "as_async", "catch_with", "catch_as_result"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ..arctrl_validation_packages.validation_package import ValidationPackage
from ..arctrl_validation_packages.validation_packages_config import ValidationPackagesConfig
from ..fable_library.list import (choose, of_array, FSharpList)
from ..fable_library.seq import (to_list, delay, collect, singleton)
from ..fable_library.types import Array
from ..fable_library.util import (equals, IEnumerable_1)
from ..yamlicious.yamlicious_types import YAMLContent_create_27AED5E3
from ..yamlicious.decode import (object, IRequiredGetter, resizearray as resizearray_1, IOptionalGetter, string as string_1, IGetters)
from ..yamlicious.encode import (try_include, string, resizearray)
from ..yamlicious.reader import read
from ..yamlicious.writer import write
from ..yamlicious.yamlicious_types import (YAMLElement, Config)
from .encode import default_whitespace
from .validation_package import (ValidationPackage_encoder, ValidationPackage_decoder)

def ValidationPackagesConfig_encoder(validationpackage: ValidationPackagesConfig) -> YAMLElement:
    def chooser(tupled_arg: tuple[str, YAMLElement], validationpackage: Any=validationpackage) -> tuple[str, YAMLElement] | None:
        v: YAMLElement = tupled_arg[1]
        if equals(v, YAMLElement(5)):
            return None

        else: 
            return (tupled_arg[0], v)


    def _arrow3677(value: str, validationpackage: Any=validationpackage) -> YAMLElement:
        return string(value)

    def _arrow3678(validationpackage_1: ValidationPackage, validationpackage: Any=validationpackage) -> YAMLElement:
        return ValidationPackage_encoder(validationpackage_1)

    obj_seq: FSharpList[tuple[str, YAMLElement]] = choose(chooser, of_array([try_include("arc_specification", _arrow3677, validationpackage.ARCSpecification), ("validation_packages", resizearray(_arrow3678, validationpackage.ValidationPackages))]))
    def _arrow3680(__unit: None=None, validationpackage: Any=validationpackage) -> IEnumerable_1[YAMLElement]:
        def _arrow3679(match_value: tuple[str, YAMLElement]) -> IEnumerable_1[YAMLElement]:
            return singleton(YAMLElement(0, YAMLContent_create_27AED5E3(match_value[0]), match_value[1]))

        return collect(_arrow3679, obj_seq)

    return YAMLElement(3, to_list(delay(_arrow3680)))


def _arrow3683(value_2: YAMLElement) -> ValidationPackagesConfig:
    def getter(get: IGetters) -> ValidationPackagesConfig:
        def _arrow3681(__unit: None=None, get: Any=get) -> Array[ValidationPackage]:
            object_arg: IRequiredGetter = get.Required
            def arg_1(value: YAMLElement) -> Array[ValidationPackage]:
                return resizearray_1(ValidationPackage_decoder, value)

            return object_arg.Field("validation_packages", arg_1)

        def _arrow3682(__unit: None=None, get: Any=get) -> str | None:
            object_arg_1: IOptionalGetter = get.Optional
            return object_arg_1.Field("arc_specification", string_1)

        return ValidationPackagesConfig(_arrow3681(), _arrow3682())

    return object(getter, value_2)


ValidationPackagesConfig_decoder: Callable[[YAMLElement], ValidationPackagesConfig] = _arrow3683

def ARCtrl_ValidationPackages_ValidationPackagesConfig__ValidationPackagesConfig_fromYamlString_Static_Z721C83C5(s: str) -> ValidationPackagesConfig:
    return ValidationPackagesConfig_decoder(read(s))


def ARCtrl_ValidationPackages_ValidationPackagesConfig__ValidationPackagesConfig_toYamlString_Static_71136F3F(whitespace: int | None=None) -> Callable[[ValidationPackagesConfig], str]:
    def _arrow3685(vp: ValidationPackagesConfig, whitespace: Any=whitespace) -> str:
        element: YAMLElement = ValidationPackagesConfig_encoder(vp)
        whitespace_1: int = default_whitespace(whitespace) or 0
        def _arrow3684(c: Config) -> Config:
            return Config(whitespace_1, c.Level)

        return write(element, _arrow3684)

    return _arrow3685


def ARCtrl_ValidationPackages_ValidationPackagesConfig__ValidationPackagesConfig_toYamlString_71136F3F(this: ValidationPackagesConfig, whitespace: int | None=None) -> str:
    return ARCtrl_ValidationPackages_ValidationPackagesConfig__ValidationPackagesConfig_toYamlString_Static_71136F3F(whitespace)(this)


__all__ = ["ValidationPackagesConfig_encoder", "ValidationPackagesConfig_decoder", "ARCtrl_ValidationPackages_ValidationPackagesConfig__ValidationPackagesConfig_fromYamlString_Static_Z721C83C5", "ARCtrl_ValidationPackages_ValidationPackagesConfig__ValidationPackagesConfig_toYamlString_Static_71136F3F", "ARCtrl_ValidationPackages_ValidationPackagesConfig__ValidationPackagesConfig_toYamlString_71136F3F"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...fable_library.array_ import (iterate as iterate_2, sort_descending)
from ...fable_library.list import (FSharpList, is_empty, iterate, append)
from ...fable_library.map_util import get_item_from_dict
from ...fable_library.option import (default_arg, map as map_1)
from ...fable_library.range import range_big_int
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.seq import (to_array, delay, map, remove_at, iterate as iterate_1, try_find_index, length, collect, singleton, choose, indexed, to_list, map_indexed, append as append_1)
from ...fable_library.string_ import (to_fail, printf, join)
from ...fable_library.system_text import (StringBuilder__ctor, StringBuilder__AppendLine_Z721C83C5)
from ...fable_library.types import (Array, to_string)
from ...fable_library.util import (ignore, IEnumerable_1, get_enumerator, compare_primitives, equals, dispose, safe_hash)
from ..Helper.collections_ import (ResizeArray_map, ResizeArray_iteri, ResizeArray_singleton, ResizeArray_iter, ResizeArray_init, ResizeArray_choose, ResizeArray_mapi, ResizeArray_groupBy, ResizeArray_indexed)
from ..Helper.hash_codes import (box_hash_array, box_hash_seq)
from ..ontology_annotation import OntologyAnnotation
from .arc_table_aux import (SanityChecks_validateArcTableValues, ArcTableValues, Unchecked_tryGetCellAt, SanityChecks_validateColumnIndex, SanityChecks_validateRowIndex, Unchecked_getCellWithDefault, Unchecked_setCellAt, try_find_duplicate_unique, SanityChecks_validateColumn, Unchecked_addColumn, Unchecked_addColumnFill, Unchecked_removeHeader, Unchecked_removeColumnCells, try_find_duplicate_unique_in_array, Unchecked_removeColumnCells_withIndexChange, ColumnValueRefs, get_empty_cell_for_header, Unchecked_moveColumnTo, Unchecked_addRow, Unchecked_addEmptyRow, Unchecked_addRows, Unchecked_removeRowCells_withIndexChange, Unchecked_alignByHeaders, SanityChecks_validateCellColumns)
from .composite_cell import CompositeCell
from .composite_column import CompositeColumn
from .composite_header import CompositeHeader

def _expr973() -> TypeInfo:
    return class_type("ARCtrl.ArcTable", None, ArcTable)


class ArcTable:
    def __init__(self, name: str, headers: Array[CompositeHeader] | None=None, columns: Array[Array[CompositeCell]] | None=None) -> None:
        headers_1: Array[CompositeHeader] = default_arg(headers, [])
        columns_1: Array[Array[CompositeCell]] = default_arg(columns, [])
        valid: bool = SanityChecks_validateCellColumns(headers_1, columns_1, True)
        self._values: ArcTableValues = ArcTableValues.from_cell_columns(columns_1)
        self._name: str = name
        self._headers: Array[CompositeHeader] = headers_1

    @property
    def ValueMap(self, __unit: None=None) -> Any:
        this: ArcTable = self
        return this._values.ValueMap

    @ValueMap.setter
    def ValueMap(self, value_map: Any) -> None:
        this: ArcTable = self
        this._values.ValueMap = value_map

    @property
    def ColumnRefs(self, __unit: None=None) -> Any:
        this: ArcTable = self
        return this._values.Columns

    @ColumnRefs.setter
    def ColumnRefs(self, internal_column_refs: Any) -> None:
        this: ArcTable = self
        this._values.Columns = internal_column_refs

    @property
    def Headers(self, __unit: None=None) -> Array[CompositeHeader]:
        this: ArcTable = self
        return this._headers

    @Headers.setter
    def Headers(self, new_headers: Array[CompositeHeader]) -> None:
        this: ArcTable = self
        ignore(SanityChecks_validateArcTableValues(new_headers, this._values, True))
        this._headers = new_headers

    @property
    def Values(self, __unit: None=None) -> ArcTableValues:
        this: ArcTable = self
        return this._values

    @property
    def Name(self, __unit: None=None) -> str:
        this: ArcTable = self
        return this._name

    @Name.setter
    def Name(self, new_name: str) -> None:
        this: ArcTable = self
        this._name = new_name

    @staticmethod
    def create(name: str, headers: Array[CompositeHeader], values: Array[Array[CompositeCell]]) -> ArcTable:
        return ArcTable(name, headers, values)

    @staticmethod
    def init(name: str) -> ArcTable:
        return ArcTable(name, [], [])

    @staticmethod
    def from_arc_table_values(name: str, headers: Array[CompositeHeader], values: ArcTableValues) -> ArcTable:
        t: ArcTable = ArcTable.init(name)
        t.Headers = headers
        t.ValueMap = values.ValueMap
        t.ColumnRefs = values.Columns
        t.RowCount = values.RowCount or 0
        return t

    @staticmethod
    def create_from_rows(name: str, headers: Array[CompositeHeader], rows: Array[Array[CompositeCell]]) -> ArcTable:
        t: ArcTable = ArcTable(name, headers)
        t.AddRows(rows)
        return t

    def Validate(self, raise_exception: bool | None=None) -> bool:
        this: ArcTable = self
        raise_exception_1: bool = default_arg(raise_exception, True)
        return SanityChecks_validateArcTableValues(this.Headers, this.Values, raise_exception_1)

    @staticmethod
    def validate(raise_exception: bool | None=None) -> Callable[[ArcTable], bool]:
        def _arrow897(table: ArcTable) -> bool:
            return table.Validate(raise_exception)

        return _arrow897

    @property
    def ColumnCount(self, __unit: None=None) -> int:
        this: ArcTable = self
        return len(this.Headers)

    @staticmethod
    def column_count(table: ArcTable) -> int:
        return table.ColumnCount

    @property
    def RowCount(self, __unit: None=None) -> int:
        this: ArcTable = self
        return this._values.RowCount

    @RowCount.setter
    def RowCount(self, new_row_count: int) -> None:
        this: ArcTable = self
        this._values.RowCount = new_row_count or 0

    @staticmethod
    def row_count(table: ArcTable) -> int:
        return table.RowCount

    @staticmethod
    def set_row_count(new_row_count: int) -> Callable[[ArcTable], ArcTable]:
        def _arrow898(table: ArcTable) -> ArcTable:
            new_table: ArcTable = table.Copy()
            new_table.RowCount = new_row_count or 0
            return new_table

        return _arrow898

    @property
    def Columns(self, __unit: None=None) -> Array[CompositeColumn]:
        this: ArcTable = self
        def _arrow901(__unit: None=None) -> IEnumerable_1[CompositeColumn]:
            def _arrow900(i: int) -> CompositeColumn:
                return this.GetColumn(i)

            return map(_arrow900, range_big_int(0, 1, this.ColumnCount - 1))

        return list(to_array(delay(_arrow901)))

    def Copy(self, __unit: None=None) -> ArcTable:
        this: ArcTable = self
        def f(h: CompositeHeader) -> CompositeHeader:
            return h.Copy()

        next_headers: Array[CompositeHeader] = ResizeArray_map(f, this.Headers)
        next_values: ArcTableValues = this._values.Copy()
        return ArcTable.from_arc_table_values(this.Name, next_headers, next_values)

    def TryGetCellAt(self, column: int, row: int) -> CompositeCell | None:
        this: ArcTable = self
        return Unchecked_tryGetCellAt(column, row, this._values)

    @staticmethod
    def try_get_cell_at(column: int, row: int) -> Callable[[ArcTable], CompositeCell | None]:
        def _arrow903(table: ArcTable) -> CompositeCell | None:
            return table.TryGetCellAt(column, row)

        return _arrow903

    def GetCellAt(self, column: int, row: int) -> CompositeCell:
        this: ArcTable = self
        SanityChecks_validateColumnIndex(column, this.ColumnCount, False)
        SanityChecks_validateRowIndex(row, this.RowCount, False)
        return Unchecked_getCellWithDefault(column, row, this._headers, this._values)

    @staticmethod
    def get_cell_at(column: int, row: int) -> Callable[[ArcTable], CompositeCell]:
        def _arrow904(table: ArcTable) -> CompositeCell:
            return table.GetCellAt(column, row)

        return _arrow904

    def IterColumns(self, action: Callable[[CompositeColumn], None]) -> None:
        this: ArcTable = self
        for column_index in range(0, (this.ColumnCount - 1) + 1, 1):
            action(this.GetColumn(column_index))

    @staticmethod
    def iter_columns(action: Callable[[CompositeColumn], None]) -> Callable[[ArcTable], ArcTable]:
        def _arrow906(table: ArcTable) -> ArcTable:
            copy: ArcTable = table.Copy()
            copy.IterColumns(action)
            return copy

        return _arrow906

    def IteriColumns(self, action: Callable[[int, CompositeColumn], None]) -> None:
        this: ArcTable = self
        for column_index in range(0, (this.ColumnCount - 1) + 1, 1):
            action(column_index, this.GetColumn(column_index))

    @staticmethod
    def iteri_columns(action: Callable[[int, CompositeColumn], None]) -> Callable[[ArcTable], ArcTable]:
        def _arrow907(table: ArcTable) -> ArcTable:
            copy: ArcTable = table.Copy()
            copy.IteriColumns(action)
            return copy

        return _arrow907

    def UpdateCellAt(self, column_index: int, row_index: int, c: CompositeCell, skip_validation: bool | None=None) -> None:
        this: ArcTable = self
        if not default_arg(skip_validation, False):
            SanityChecks_validateColumnIndex(column_index, this.ColumnCount, False)
            SanityChecks_validateRowIndex(row_index, this.RowCount, False)
            ignore(c.ValidateAgainstHeader(this.Headers[column_index], True))

        Unchecked_setCellAt(column_index, row_index, c, this._values)

    @staticmethod
    def update_cell_at(column_index: int, row_index: int, cell: CompositeCell, skip_validation: bool | None=None) -> Callable[[ArcTable], ArcTable]:
        def _arrow909(table: ArcTable) -> ArcTable:
            new_table: ArcTable = table.Copy()
            new_table.UpdateCellAt(column_index, row_index, cell, skip_validation)
            return new_table

        return _arrow909

    def SetCellAt(self, column_index: int, row_index: int, c: CompositeCell, skip_validation: bool | None=None) -> None:
        this: ArcTable = self
        if not default_arg(skip_validation, False):
            SanityChecks_validateColumnIndex(column_index, this.ColumnCount, False)
            ignore(c.ValidateAgainstHeader(this.Headers[column_index], True))

        Unchecked_setCellAt(column_index, row_index, c, this._values)

    @staticmethod
    def set_cell_at(column_index: int, row_index: int, cell: CompositeCell, skip_validation: bool | None=None) -> Callable[[ArcTable], ArcTable]:
        def _arrow911(table: ArcTable) -> ArcTable:
            new_table: ArcTable = table.Copy()
            new_table.SetCellAt(column_index, row_index, cell, skip_validation)
            return new_table

        return _arrow911

    def UpdateCellsBy(self, f: Callable[[int, int, CompositeCell], CompositeCell], skip_validation: bool | None=None) -> None:
        this: ArcTable = self
        skip_validation_1: bool = default_arg(skip_validation, False)
        with get_enumerator(this._values) as enumerator:
            while enumerator.System_Collections_IEnumerator_MoveNext():
                kv: Any = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
                pattern_input: tuple[int, int] = kv[0]
                ri: int = pattern_input[1] or 0
                ci: int = pattern_input[0] or 0
                new_cell: CompositeCell = f(ci, ri, kv[1])
                if not skip_validation_1:
                    ignore(new_cell.ValidateAgainstHeader(this.Headers[ci], True))

                Unchecked_setCellAt(ci, ri, new_cell, this._values)

    @staticmethod
    def update_cells_by(f: Callable[[int, int, CompositeCell], CompositeCell], skip_validation: bool | None=None) -> Callable[[ArcTable], None]:
        def _arrow913(table: ArcTable) -> None:
            new_table: ArcTable = table.Copy()
            new_table.UpdateCellsBy(f, skip_validation)

        return _arrow913

    def UpdateCellBy(self, column_index: int, row_index: int, f: Callable[[CompositeCell], CompositeCell], skip_validation: bool | None=None) -> None:
        this: ArcTable = self
        skip_validation_1: bool = default_arg(skip_validation, False)
        if not skip_validation_1:
            SanityChecks_validateColumnIndex(column_index, this.ColumnCount, False)
            SanityChecks_validateRowIndex(row_index, this.RowCount, False)

        new_cell: CompositeCell = f(this.GetCellAt(column_index, row_index))
        if not skip_validation_1:
            ignore(new_cell.ValidateAgainstHeader(this.Headers[column_index], True))

        Unchecked_setCellAt(column_index, row_index, new_cell, this._values)

    @staticmethod
    def update_cell_by(column_index: int, row_index: int, f: Callable[[CompositeCell], CompositeCell], skip_validation: bool | None=None) -> Callable[[ArcTable], None]:
        def _arrow914(table: ArcTable) -> None:
            new_table: ArcTable = table.Copy()
            new_table.UpdateCellBy(column_index, row_index, f, skip_validation)

        return _arrow914

    def UpdateHeader(self, index: int, new_header: CompositeHeader, force_convert_cells: bool | None=None) -> None:
        this: ArcTable = self
        force_convert_cells_1: bool = default_arg(force_convert_cells, False)
        SanityChecks_validateColumnIndex(index, this.ColumnCount, False)
        header: CompositeHeader = new_header
        match_value: int | None = try_find_duplicate_unique(header, remove_at(index, this.Headers))
        if match_value is not None:
            raise Exception(((("Invalid input. Tried setting unique header `" + str(header)) + "`, but header of same type already exists at index ") + str(match_value)) + ".")

        c: CompositeColumn = CompositeColumn(new_header, this.GetColumn(index).Cells)
        if c.Validate():
            set_header: None
            this.Headers[index] = new_header

        elif force_convert_cells_1:
            def f(c_1: CompositeCell) -> CompositeCell:
                if c_1.is_free_text:
                    return c_1.ToTermCell()

                else: 
                    return c_1


            def f_1(c_2: CompositeCell) -> CompositeCell:
                return c_2.ToFreeTextCell()

            converted_cells: Array[CompositeCell] = ResizeArray_map(f, c.Cells) if new_header.IsTermColumn else ResizeArray_map(f_1, c.Cells)
            this.UpdateColumn(index, new_header, converted_cells)

        else: 
            raise Exception("Tried setting header for column with invalid type of cells. Set `forceConvertCells` flag to automatically convert cells into valid CompositeCell type.")


    @staticmethod
    def update_header(index: int, header: CompositeHeader) -> Callable[[ArcTable], ArcTable]:
        def _arrow918(table: ArcTable) -> ArcTable:
            new_table: ArcTable = table.Copy()
            new_table.UpdateHeader(index, header)
            return new_table

        return _arrow918

    def AddColumn(self, header: CompositeHeader, cells: Array[CompositeCell] | None=None, index: int | None=None, force_replace: bool | None=None) -> None:
        this: ArcTable = self
        index_1: int = default_arg(index, this.ColumnCount) or 0
        cells_1: Array[CompositeCell] = default_arg(cells, [])
        force_replace_1: bool = default_arg(force_replace, False)
        SanityChecks_validateColumnIndex(index_1, this.ColumnCount, True)
        SanityChecks_validateColumn(CompositeColumn.create(header, cells_1))
        Unchecked_addColumn(header, cells_1, index_1, force_replace_1, False, this.Headers, this._values)

    @staticmethod
    def add_column(header: CompositeHeader, cells: Array[CompositeCell] | None=None, index: int | None=None, force_replace: bool | None=None) -> Callable[[ArcTable], ArcTable]:
        def _arrow921(table: ArcTable) -> ArcTable:
            new_table: ArcTable = table.Copy()
            new_table.AddColumn(header, cells, index, force_replace)
            return new_table

        return _arrow921

    def AddColumnFill(self, header: CompositeHeader, cell: CompositeCell, index: int | None=None, force_replace: bool | None=None) -> None:
        this: ArcTable = self
        index_1: int = default_arg(index, this.ColumnCount) or 0
        force_replace_1: bool = default_arg(force_replace, False)
        SanityChecks_validateColumnIndex(index_1, this.ColumnCount, True)
        SanityChecks_validateColumn(CompositeColumn.create(header, [cell]))
        Unchecked_addColumnFill(header, cell, index_1, force_replace_1, False, this.Headers, this._values)

    @staticmethod
    def add_column_fill(header: CompositeHeader, cell: CompositeCell, index: int | None=None, force_replace: bool | None=None) -> Callable[[ArcTable], ArcTable]:
        def _arrow922(table: ArcTable) -> ArcTable:
            new_table: ArcTable = table.Copy()
            new_table.AddColumnFill(header, cell, index, force_replace)
            return new_table

        return _arrow922

    def UpdateColumn(self, column_index: int, header: CompositeHeader, cells: Array[CompositeCell] | None=None) -> None:
        this: ArcTable = self
        SanityChecks_validateColumnIndex(column_index, this.ColumnCount, False)
        column: CompositeColumn = CompositeColumn.create(header, cells)
        SanityChecks_validateColumn(column)
        header_1: CompositeHeader = column.Header
        match_value: int | None = try_find_duplicate_unique(header_1, remove_at(column_index, this.Headers))
        if match_value is not None:
            raise Exception(((("Invalid input. Tried setting unique header `" + str(header_1)) + "`, but header of same type already exists at index ") + str(match_value)) + ".")

        Unchecked_removeHeader(column_index, this.Headers)
        Unchecked_removeColumnCells(column_index, this._values)
        this.Headers.insert(column_index, column.Header)
        def f(row_index: int, v: CompositeCell) -> None:
            Unchecked_setCellAt(column_index, row_index, v, this._values)

        ResizeArray_iteri(f, column.Cells)

    @staticmethod
    def update_column(column_index: int, header: CompositeHeader, cells: Array[CompositeCell] | None=None) -> Callable[[ArcTable], ArcTable]:
        def _arrow923(table: ArcTable) -> ArcTable:
            new_table: ArcTable = table.Copy()
            new_table.UpdateColumn(column_index, header, cells)
            return new_table

        return _arrow923

    def InsertColumn(self, index: int, header: CompositeHeader, cells: Array[CompositeCell] | None=None) -> None:
        this: ArcTable = self
        this.AddColumn(header, cells, index, False)

    @staticmethod
    def insert_column(index: int, header: CompositeHeader, cells: Array[CompositeCell] | None=None) -> Callable[[ArcTable], ArcTable]:
        def _arrow924(table: ArcTable) -> ArcTable:
            new_table: ArcTable = table.Copy()
            new_table.InsertColumn(index, header, cells)
            return new_table

        return _arrow924

    def AppendColumn(self, header: CompositeHeader, cells: Array[CompositeCell] | None=None) -> None:
        this: ArcTable = self
        this.AddColumn(header, cells, this.ColumnCount, False)

    @staticmethod
    def append_column(header: CompositeHeader, cells: Array[CompositeCell] | None=None) -> Callable[[ArcTable], ArcTable]:
        def _arrow925(table: ArcTable) -> ArcTable:
            new_table: ArcTable = table.Copy()
            new_table.AppendColumn(header, cells)
            return new_table

        return _arrow925

    def AddColumns(self, columns: IEnumerable_1[CompositeColumn], index: int | None=None, force_replace: bool | None=None) -> None:
        this: ArcTable = self
        index_1: int = default_arg(index, this.ColumnCount) or 0
        force_replace_1: bool = default_arg(force_replace, False)
        SanityChecks_validateColumnIndex(index_1, this.ColumnCount, True)
        def mapping(x: CompositeColumn) -> CompositeHeader:
            return x.Header

        duplicates: FSharpList[dict[str, Any]] = try_find_duplicate_unique_in_array(map(mapping, columns))
        if not is_empty(duplicates):
            sb: Any = StringBuilder__ctor()
            ignore(StringBuilder__AppendLine_Z721C83C5(sb, "Found duplicate unique columns in `columns`."))
            def action(x_1: dict[str, Any]) -> None:
                ignore(StringBuilder__AppendLine_Z721C83C5(sb, ((((("Duplicate `" + str(x_1["HeaderType"])) + "` at index ") + str(x_1["Index1"])) + " and ") + str(x_1["Index2"])) + "."))

            iterate(action, duplicates)
            raise Exception(to_string(sb))

        def action_1(x_2: CompositeColumn) -> None:
            SanityChecks_validateColumn(x_2)

        iterate_1(action_1, columns)
        def action_2(col: CompositeColumn) -> None:
            nonlocal index_1
            prev_headers_count: int = len(this.Headers) or 0
            Unchecked_addColumn(col.Header, col.Cells, index_1, force_replace_1, False, this.Headers, this._values)
            if len(this.Headers) > prev_headers_count:
                index_1 = (index_1 + 1) or 0


        iterate_1(action_2, columns)

    @staticmethod
    def add_columns(columns: Array[CompositeColumn], index: int | None=None) -> Callable[[ArcTable], ArcTable]:
        def _arrow926(table: ArcTable) -> ArcTable:
            new_table: ArcTable = table.Copy()
            new_table.AddColumns(columns, index)
            return new_table

        return _arrow926

    def RemoveColumn(self, index: int) -> None:
        this: ArcTable = self
        SanityChecks_validateColumnIndex(index, this.ColumnCount, False)
        column_count: int = this.ColumnCount or 0
        Unchecked_removeHeader(index, this.Headers)
        Unchecked_removeColumnCells_withIndexChange(index, column_count, this._values)

    @staticmethod
    def remove_column(index: int) -> Callable[[ArcTable], ArcTable]:
        def _arrow927(table: ArcTable) -> ArcTable:
            new_table: ArcTable = table.Copy()
            new_table.RemoveColumn(index)
            return new_table

        return _arrow927

    def RemoveColumns(self, index_arr: Array[int]) -> None:
        this: ArcTable = self
        def _arrow928(index: int) -> None:
            SanityChecks_validateColumnIndex(index, this.ColumnCount, False)

        iterate_2(_arrow928, index_arr)
        def _arrow929(index_1: int) -> None:
            this.RemoveColumn(index_1)

        class ObjectExpr930:
            @property
            def Compare(self) -> Callable[[int, int], int]:
                return compare_primitives

        iterate_2(_arrow929, sort_descending(index_arr, ObjectExpr930()))

    @staticmethod
    def remove_columns(index_arr: Array[int]) -> Callable[[ArcTable], ArcTable]:
        def _arrow931(table: ArcTable) -> ArcTable:
            new_table: ArcTable = table.Copy()
            new_table.RemoveColumns(index_arr)
            return new_table

        return _arrow931

    def GetColumn(self, column_index: int, fail_on_missing_cell: bool | None=None) -> CompositeColumn:
        this: ArcTable = self
        fail_on_missing_cell_1: bool = default_arg(fail_on_missing_cell, False)
        SanityChecks_validateColumnIndex(column_index, this.ColumnCount, False)
        h: CompositeHeader = this.Headers[column_index]
        cells: Array[CompositeCell] = []
        if this.RowCount != 0:
            col: ColumnValueRefs = get_item_from_dict(this._values.Columns, column_index)
            if col.tag == 1:
                vals: Any = col.fields[0]
                empty_cell: CompositeCell = get_empty_cell_for_header(h, None)
                for row_index in range(0, (this.RowCount - 1) + 1, 1):
                    if row_index in vals:
                        (cells.append(get_item_from_dict(this._values.ValueMap, get_item_from_dict(vals, row_index))))

                    elif fail_on_missing_cell_1:
                        to_fail(printf("Could not return column: Unable to find cell for index: (%i, %i)"))(column_index)(row_index)

                    else: 
                        (cells.append(empty_cell))


            else: 
                c: CompositeCell = get_item_from_dict(this._values.ValueMap, col.fields[0])
                for i in range(0, (this.RowCount - 1) + 1, 1):
                    (cells.append(c))


        return CompositeColumn.create(h, cells)

    @staticmethod
    def get_column(index: int, fail_on_missing_cell: bool | None=None) -> Callable[[ArcTable], CompositeColumn]:
        def _arrow932(table: ArcTable) -> CompositeColumn:
            return table.GetColumn(index, fail_on_missing_cell)

        return _arrow932

    def TryGetColumnByHeader(self, header: CompositeHeader, fail_on_missing_cell: bool | None=None) -> CompositeColumn | None:
        this: ArcTable = self
        def mapping(i: int) -> CompositeColumn:
            return this.GetColumn(i, fail_on_missing_cell)

        def predicate(x: CompositeHeader) -> bool:
            return equals(x, header)

        return map_1(mapping, try_find_index(predicate, this.Headers))

    @staticmethod
    def try_get_column_by_header(header: CompositeHeader, fail_on_missing_cell: bool | None=None) -> Callable[[ArcTable], CompositeColumn | None]:
        def _arrow933(table: ArcTable) -> CompositeColumn | None:
            return table.TryGetColumnByHeader(header, fail_on_missing_cell)

        return _arrow933

    def TryGetColumnByHeaderBy(self, header_predicate: Callable[[CompositeHeader], bool], fail_on_missing_cell: bool | None=None) -> CompositeColumn | None:
        this: ArcTable = self
        def mapping(i: int) -> CompositeColumn:
            return this.GetColumn(i, fail_on_missing_cell)

        return map_1(mapping, try_find_index(header_predicate, this.Headers))

    @staticmethod
    def try_get_column_by_header_by(header_predicate: Callable[[CompositeHeader], bool], fail_on_missing_cell: bool | None=None) -> Callable[[ArcTable], CompositeColumn | None]:
        def _arrow934(table: ArcTable) -> CompositeColumn | None:
            return table.TryGetColumnByHeaderBy(header_predicate, fail_on_missing_cell)

        return _arrow934

    def GetColumnByHeader(self, header: CompositeHeader, fail_on_missing_cell: bool | None=None) -> CompositeColumn:
        this: ArcTable = self
        match_value: CompositeColumn | None = this.TryGetColumnByHeader(header, fail_on_missing_cell)
        if match_value is None:
            arg: str = this.Name
            return to_fail(printf("Unable to find column with header in table %s: %O"))(arg)(header)

        else: 
            return match_value


    @staticmethod
    def get_column_by_header(header: CompositeHeader, fail_on_missing_cell: bool | None=None) -> Callable[[ArcTable], CompositeColumn]:
        def _arrow935(table: ArcTable) -> CompositeColumn:
            return table.GetColumnByHeader(header, fail_on_missing_cell)

        return _arrow935

    def TryGetInputColumn(self, __unit: None=None) -> CompositeColumn | None:
        this: ArcTable = self
        def mapping(i: int) -> CompositeColumn:
            return this.GetColumn(i)

        def predicate(x: CompositeHeader) -> bool:
            return x.is_input

        return map_1(mapping, try_find_index(predicate, this.Headers))

    @staticmethod
    def try_get_input_column(__unit: None=None) -> Callable[[ArcTable], CompositeColumn | None]:
        def _arrow936(table: ArcTable) -> CompositeColumn | None:
            return table.TryGetInputColumn()

        return _arrow936

    def GetInputColumn(self, __unit: None=None) -> CompositeColumn:
        this: ArcTable = self
        match_value: CompositeColumn | None = this.TryGetInputColumn()
        if match_value is None:
            arg: str = this.Name
            return to_fail(printf("Unable to find input column in table %s"))(arg)

        else: 
            return match_value


    @staticmethod
    def get_input_column(__unit: None=None) -> Callable[[ArcTable], CompositeColumn]:
        def _arrow937(table: ArcTable) -> CompositeColumn:
            return table.GetInputColumn()

        return _arrow937

    def TryGetOutputColumn(self, __unit: None=None) -> CompositeColumn | None:
        this: ArcTable = self
        def mapping(i: int) -> CompositeColumn:
            return this.GetColumn(i)

        def predicate(x: CompositeHeader) -> bool:
            return x.is_output

        return map_1(mapping, try_find_index(predicate, this.Headers))

    @staticmethod
    def try_get_output_column(__unit: None=None) -> Callable[[ArcTable], CompositeColumn | None]:
        def _arrow938(table: ArcTable) -> CompositeColumn | None:
            return table.TryGetOutputColumn()

        return _arrow938

    def GetOutputColumn(self, __unit: None=None) -> CompositeColumn:
        this: ArcTable = self
        match_value: CompositeColumn | None = this.TryGetOutputColumn()
        if match_value is None:
            arg: str = this.Name
            return to_fail(printf("Unable to find output column in table %s"))(arg)

        else: 
            return match_value


    @staticmethod
    def get_output_column(__unit: None=None) -> Callable[[ArcTable], CompositeColumn]:
        def _arrow939(table: ArcTable) -> CompositeColumn:
            return table.GetOutputColumn()

        return _arrow939

    def MoveColumn(self, start_col: int, end_col: int) -> None:
        this: ArcTable = self
        if start_col == end_col:
            pass

        elif True if (start_col < 0) else (start_col >= this.ColumnCount):
            to_fail(printf("Cannt move column. Invalid start column index: %i"))(start_col)

        elif True if (end_col < 0) else (end_col >= this.ColumnCount):
            to_fail(printf("Cannt move column. Invalid end column index: %i"))(end_col)

        else: 
            Unchecked_moveColumnTo(start_col, end_col, this.Headers, this._values)


    @staticmethod
    def move_column(start_col: int, end_col: int) -> Callable[[ArcTable], ArcTable]:
        def _arrow940(table: ArcTable) -> ArcTable:
            new_table: ArcTable = table.Copy()
            new_table.MoveColumn(start_col, end_col)
            return new_table

        return _arrow940

    def AddRow(self, cells: Array[CompositeCell] | None=None, index: int | None=None) -> None:
        this: ArcTable = self
        index_1: int = default_arg(index, this.RowCount) or 0
        SanityChecks_validateRowIndex(index_1, this.RowCount, True)
        if cells is not None:
            cells_1: Array[CompositeCell] = cells
            column_count: int = this.ColumnCount or 0
            new_cells_count: int = length(cells_1) or 0
            if column_count == 0:
                raise Exception("Table contains no columns! Cannot add row to empty table!")

            elif new_cells_count != column_count:
                raise Exception(((("Cannot add a new row with " + str(new_cells_count)) + " cells, as the table has ") + str(column_count)) + " columns.")

            for column_index in range(0, (this.ColumnCount - 1) + 1, 1):
                h: CompositeHeader = this.Headers[column_index]
                column: CompositeColumn = CompositeColumn.create(h, [cells_1[column_index]])
                ignore(cells_1[column_index].ValidateAgainstHeader(h, True))
            Unchecked_addRow(index_1, cells_1, this.Headers, this._values)

        else: 
            Unchecked_addEmptyRow(index_1, this._values)


    @staticmethod
    def add_row(cells: Array[CompositeCell] | None=None, index: int | None=None) -> Callable[[ArcTable], ArcTable]:
        def _arrow941(table: ArcTable) -> ArcTable:
            new_table: ArcTable = table.Copy()
            new_table.AddRow(cells, index)
            return new_table

        return _arrow941

    def UpdateRow(self, row_index: int, cells: Array[CompositeCell]) -> None:
        this: ArcTable = self
        SanityChecks_validateRowIndex(row_index, this.RowCount, False)
        column_count: int = this.RowCount or 0
        new_cells_count: int = length(cells) or 0
        if column_count == 0:
            raise Exception("Table contains no columns! Cannot add row to empty table!")

        elif new_cells_count != column_count:
            raise Exception(((("Cannot add a new row with " + str(new_cells_count)) + " cells, as the table has ") + str(column_count)) + " columns.")

        def f(i: int, cell: CompositeCell) -> None:
            h: CompositeHeader = this.Headers[i]
            SanityChecks_validateColumn(CompositeColumn.create(h, ResizeArray_singleton(cell)))

        ResizeArray_iteri(f, cells)
        def f_1(column_index: int, cell_1: CompositeCell) -> None:
            Unchecked_setCellAt(column_index, row_index, cell_1, this._values)

        ResizeArray_iteri(f_1, cells)

    @staticmethod
    def update_row(row_index: int, cells: Array[CompositeCell]) -> Callable[[ArcTable], ArcTable]:
        def _arrow942(table: ArcTable) -> ArcTable:
            new_table: ArcTable = table.Copy()
            new_table.UpdateRow(row_index, cells)
            return new_table

        return _arrow942

    def AppendRow(self, cells: Array[CompositeCell] | None=None) -> None:
        this: ArcTable = self
        this.AddRow(cells, this.RowCount)

    @staticmethod
    def append_row(cells: Array[CompositeCell] | None=None) -> Callable[[ArcTable], ArcTable]:
        def _arrow943(table: ArcTable) -> ArcTable:
            new_table: ArcTable = table.Copy()
            new_table.AppendRow(cells)
            return new_table

        return _arrow943

    def InsertRow(self, index: int, cells: Array[CompositeCell] | None=None) -> None:
        this: ArcTable = self
        this.AddRow(cells, index)

    @staticmethod
    def insert_row(index: int, cells: Array[CompositeCell] | None=None) -> Callable[[ArcTable], ArcTable]:
        def _arrow944(table: ArcTable) -> ArcTable:
            new_table: ArcTable = table.Copy()
            new_table.AddRow(cells, index)
            return new_table

        return _arrow944

    def AddRows(self, rows: Array[Array[CompositeCell]], index: int | None=None) -> None:
        this: ArcTable = self
        index_1: int = default_arg(index, this.RowCount) or 0
        SanityChecks_validateRowIndex(index_1, this.RowCount, True)
        def f(row: Array[CompositeCell]) -> None:
            column_count: int = this.ColumnCount or 0
            new_cells_count: int = length(row) or 0
            if column_count == 0:
                raise Exception("Table contains no columns! Cannot add row to empty table!")

            elif new_cells_count != column_count:
                raise Exception(((("Cannot add a new row with " + str(new_cells_count)) + " cells, as the table has ") + str(column_count)) + " columns.")


        ResizeArray_iter(f, rows)
        enumerator: Any = get_enumerator(rows)
        try: 
            while enumerator.System_Collections_IEnumerator_MoveNext():
                row_1: Array[CompositeCell] = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
                for column_index in range(0, (this.ColumnCount - 1) + 1, 1):
                    h: CompositeHeader = this.Headers[column_index]
                    ignore(row_1[column_index].ValidateAgainstHeader(h, True))

        finally: 
            dispose(enumerator)

        Unchecked_addRows(index_1, rows, this.Headers, this._values)

    @staticmethod
    def add_rows(rows: Array[Array[CompositeCell]], index: int | None=None) -> Callable[[ArcTable], ArcTable]:
        def _arrow945(table: ArcTable) -> ArcTable:
            new_table: ArcTable = table.Copy()
            new_table.AddRows(rows, index)
            return new_table

        return _arrow945

    def AddRowsEmpty(self, row_count: int, index: int | None=None) -> None:
        this: ArcTable = self
        def _arrow947(__unit: None=None) -> IEnumerable_1[CompositeCell]:
            def _arrow946(column_index: int) -> IEnumerable_1[CompositeCell]:
                return singleton(get_empty_cell_for_header(this.Headers[column_index], Unchecked_tryGetCellAt(column_index, 0, this._values)))

            return collect(_arrow946, range_big_int(0, 1, this.ColumnCount - 1))

        row: Array[CompositeCell] = list(to_array(delay(_arrow947)))
        def _arrow948(_arg: int) -> Array[CompositeCell]:
            return row

        rows: Array[Array[CompositeCell]] = ResizeArray_init(row_count, _arrow948)
        this.AddRows(rows, index)

    @staticmethod
    def add_rows_empty(row_count: int, index: int | None=None) -> Callable[[ArcTable], ArcTable]:
        def _arrow949(table: ArcTable) -> ArcTable:
            new_table: ArcTable = table.Copy()
            new_table.AddRowsEmpty(row_count, index)
            return new_table

        return _arrow949

    def RemoveRow(self, index: int) -> None:
        this: ArcTable = self
        SanityChecks_validateRowIndex(index, this.RowCount, False)
        Unchecked_removeRowCells_withIndexChange(index, this._values)

    @staticmethod
    def remove_row(index: int) -> Callable[[ArcTable], ArcTable]:
        def _arrow950(table: ArcTable) -> ArcTable:
            new_table: ArcTable = table.Copy()
            new_table.RemoveRow(index)
            return new_table

        return _arrow950

    def RemoveRows(self, index_arr: Array[int]) -> None:
        this: ArcTable = self
        def _arrow951(index: int) -> None:
            SanityChecks_validateRowIndex(index, this.RowCount, False)

        iterate_2(_arrow951, index_arr)
        def _arrow952(index_1: int) -> None:
            this.RemoveRow(index_1)

        class ObjectExpr953:
            @property
            def Compare(self) -> Callable[[int, int], int]:
                return compare_primitives

        iterate_2(_arrow952, sort_descending(index_arr, ObjectExpr953()))

    @staticmethod
    def remove_rows(index_arr: Array[int]) -> Callable[[ArcTable], ArcTable]:
        def _arrow954(table: ArcTable) -> ArcTable:
            new_table: ArcTable = table.Copy()
            new_table.RemoveColumns(index_arr)
            return new_table

        return _arrow954

    def GetRow(self, row_index: int, SkipValidation: bool | None=None) -> Array[CompositeCell]:
        this: ArcTable = self
        if not equals(SkipValidation, True):
            SanityChecks_validateRowIndex(row_index, this.RowCount, False)

        def _arrow956(__unit: None=None) -> IEnumerable_1[CompositeCell]:
            def _arrow955(column_index: int) -> CompositeCell:
                return this.GetCellAt(column_index, row_index)

            return map(_arrow955, range_big_int(0, 1, this.ColumnCount - 1))

        return list(to_array(delay(_arrow956)))

    @staticmethod
    def get_row(index: int) -> Callable[[ArcTable], Array[CompositeCell]]:
        def _arrow957(table: ArcTable) -> Array[CompositeCell]:
            return table.GetRow(index)

        return _arrow957

    def Join(self, table: ArcTable, index: int | None=None, join_options: str | None=None, force_replace: bool | None=None) -> None:
        this: ArcTable = self
        join_options_1: str = default_arg(join_options, "headers")
        force_replace_1: bool = default_arg(force_replace, False)
        index_1: int = default_arg(index, this.ColumnCount) or 0
        index_1 = (this.ColumnCount if (index_1 == -1) else index_1) or 0
        SanityChecks_validateColumnIndex(index_1, this.ColumnCount, True)
        only_headers: bool = join_options_1 == "headers"
        columns: Array[CompositeColumn]
        pre: Array[CompositeColumn] = table.Columns
        def f_2(c_1: CompositeColumn) -> CompositeColumn:
            units_opt: Array[OntologyAnnotation] | None = c_1.TryGetColumnUnits()
            if units_opt is None:
                return CompositeColumn(c_1.Header, [])

            else: 
                def f_1(u: OntologyAnnotation, c_1: Any=c_1) -> CompositeCell:
                    return CompositeCell.create_unitized("", u)

                return CompositeColumn(c_1.Header, ResizeArray_map(f_1, units_opt))


        def f(c: CompositeColumn) -> CompositeColumn:
            return CompositeColumn(c.Header, [])

        columns = ResizeArray_map(f_2, pre) if (join_options_1 == "withUnit") else (pre if (join_options_1 == "withValues") else ResizeArray_map(f, pre))
        def mapping(x: CompositeColumn) -> CompositeHeader:
            return x.Header

        duplicates: FSharpList[dict[str, Any]] = try_find_duplicate_unique_in_array(map(mapping, columns))
        if not is_empty(duplicates):
            sb: Any = StringBuilder__ctor()
            ignore(StringBuilder__AppendLine_Z721C83C5(sb, "Found duplicate unique columns in `columns`."))
            def action(x_1: dict[str, Any]) -> None:
                ignore(StringBuilder__AppendLine_Z721C83C5(sb, ((((("Duplicate `" + str(x_1["HeaderType"])) + "` at index ") + str(x_1["Index1"])) + " and ") + str(x_1["Index2"])) + "."))

            iterate(action, duplicates)
            raise Exception(to_string(sb))

        def f_3(x_2: CompositeColumn) -> None:
            SanityChecks_validateColumn(x_2)

        ResizeArray_iter(f_3, columns)
        def f_4(col: CompositeColumn) -> None:
            nonlocal index_1
            prev_headers_count: int = len(this.Headers) or 0
            Unchecked_addColumn(col.Header, col.Cells, index_1, force_replace_1, only_headers, this.Headers, this._values)
            if len(this.Headers) > prev_headers_count:
                index_1 = (index_1 + 1) or 0


        ResizeArray_iter(f_4, columns)

    @staticmethod
    def join(table: ArcTable, index: int | None=None, join_options: str | None=None, force_replace: bool | None=None) -> Callable[[ArcTable], ArcTable]:
        def _arrow958(this: ArcTable) -> ArcTable:
            copy: ArcTable = this.Copy()
            copy.Join(table, index, join_options, force_replace)
            return copy

        return _arrow958

    def AddProtocolTypeColumn(self, types: Array[OntologyAnnotation] | None=None, index: int | None=None, force_replace: bool | None=None) -> None:
        this: ArcTable = self
        def mapping(a: Array[OntologyAnnotation]) -> Array[CompositeCell]:
            def f(Item: OntologyAnnotation, a: Any=a) -> CompositeCell:
                return CompositeCell(0, Item)

            return ResizeArray_map(f, a)

        cells: Array[CompositeCell] | None = map_1(mapping, types)
        this.AddColumn(CompositeHeader(4), cells, index, force_replace)

    def AddProtocolVersionColumn(self, versions: Array[str] | None=None, index: int | None=None, force_replace: bool | None=None) -> None:
        this: ArcTable = self
        def mapping(a: Array[str]) -> Array[CompositeCell]:
            def f(Item: str, a: Any=a) -> CompositeCell:
                return CompositeCell(1, Item)

            return ResizeArray_map(f, a)

        cells: Array[CompositeCell] | None = map_1(mapping, versions)
        this.AddColumn(CompositeHeader(7), cells, index, force_replace)

    def AddProtocolUriColumn(self, uris: Array[str] | None=None, index: int | None=None, force_replace: bool | None=None) -> None:
        this: ArcTable = self
        def mapping(a: Array[str]) -> Array[CompositeCell]:
            def f(Item: str, a: Any=a) -> CompositeCell:
                return CompositeCell(1, Item)

            return ResizeArray_map(f, a)

        cells: Array[CompositeCell] | None = map_1(mapping, uris)
        this.AddColumn(CompositeHeader(6), cells, index, force_replace)

    def AddProtocolDescriptionColumn(self, descriptions: Array[str] | None=None, index: int | None=None, force_replace: bool | None=None) -> None:
        this: ArcTable = self
        def mapping(a: Array[str]) -> Array[CompositeCell]:
            def f(Item: str, a: Any=a) -> CompositeCell:
                return CompositeCell(1, Item)

            return ResizeArray_map(f, a)

        cells: Array[CompositeCell] | None = map_1(mapping, descriptions)
        this.AddColumn(CompositeHeader(5), cells, index, force_replace)

    def AddProtocolNameColumn(self, names: Array[str] | None=None, index: int | None=None, force_replace: bool | None=None) -> None:
        this: ArcTable = self
        def mapping(a: Array[str]) -> Array[CompositeCell]:
            def f(Item: str, a: Any=a) -> CompositeCell:
                return CompositeCell(1, Item)

            return ResizeArray_map(f, a)

        cells: Array[CompositeCell] | None = map_1(mapping, names)
        this.AddColumn(CompositeHeader(8), cells, index, force_replace)

    def GetProtocolTypeColumn(self, __unit: None=None) -> CompositeColumn:
        this: ArcTable = self
        return this.GetColumnByHeader(CompositeHeader(4))

    def GetProtocolVersionColumn(self, __unit: None=None) -> CompositeColumn:
        this: ArcTable = self
        return this.GetColumnByHeader(CompositeHeader(7))

    def GetProtocolUriColumn(self, __unit: None=None) -> CompositeColumn:
        this: ArcTable = self
        return this.GetColumnByHeader(CompositeHeader(6))

    def GetProtocolDescriptionColumn(self, __unit: None=None) -> CompositeColumn:
        this: ArcTable = self
        return this.GetColumnByHeader(CompositeHeader(5))

    def GetProtocolNameColumn(self, __unit: None=None) -> CompositeColumn:
        this: ArcTable = self
        return this.GetColumnByHeader(CompositeHeader(8))

    def TryGetProtocolNameColumn(self, __unit: None=None) -> CompositeColumn | None:
        this: ArcTable = self
        return this.TryGetColumnByHeader(CompositeHeader(8))

    def GetComponentColumns(self, __unit: None=None) -> Array[CompositeColumn]:
        this: ArcTable = self
        def f(h: CompositeHeader) -> CompositeColumn | None:
            if h.is_component:
                return this.GetColumnByHeader(h)

            else: 
                return None


        return ResizeArray_choose(f, this.Headers)

    def RescanValueMap(self, __unit: None=None) -> None:
        this: ArcTable = self
        this._values.RescanValueMap()

    @staticmethod
    def SplitByColumnValues(column_index: int) -> Callable[[ArcTable], Array[ArcTable]]:
        def _arrow959(table: ArcTable) -> Array[ArcTable]:
            def f_4(i: int, index_group: Array[int]) -> ArcTable:
                headers: Array[CompositeHeader] = table.Headers
                def f_3(i_1: int, i: Any=i, index_group: Any=index_group) -> Array[CompositeCell]:
                    return table.GetRow(i_1, True)

                rows: Array[Array[CompositeCell]] = ResizeArray_map(f_3, index_group)
                return ArcTable.create_from_rows(table.Name, headers, rows)

            def f_2(tupled_arg: tuple[CompositeCell, Array[tuple[int, CompositeCell]]]) -> Array[int]:
                def f_1(tuple_1: tuple[int, CompositeCell], tupled_arg: Any=tupled_arg) -> int:
                    return tuple_1[0]

                return ResizeArray_map(f_1, tupled_arg[1])

            def f(tuple: tuple[int, CompositeCell]) -> CompositeCell:
                return tuple[1]

            return ResizeArray_mapi(f_4, ResizeArray_map(f_2, ResizeArray_groupBy(f, ResizeArray_indexed(table.GetColumn(column_index).Cells))))

        return _arrow959

    @staticmethod
    def SplitByColumnValuesByHeader(header: CompositeHeader) -> Callable[[ArcTable], Array[ArcTable]]:
        def _arrow960(table: ArcTable) -> Array[ArcTable]:
            def predicate(x: CompositeHeader) -> bool:
                return equals(x, header)

            index: int | None = try_find_index(predicate, table.Headers)
            if index is None:
                return ResizeArray_singleton(table.Copy())

            else: 
                i: int = index or 0
                return ArcTable.SplitByColumnValues(i)(table)


        return _arrow960

    @staticmethod
    def SplitByProtocolREF() -> Callable[[ArcTable], Array[ArcTable]]:
        def _arrow961(table: ArcTable) -> Array[ArcTable]:
            return ArcTable.SplitByColumnValuesByHeader(CompositeHeader(8))(table)

        return _arrow961

    @staticmethod
    def update_reference_by_annotation_table(ref_table: ArcTable, annotation_table: ArcTable) -> ArcTable:
        ref_table_1: ArcTable = ref_table.Copy()
        annotation_table_1: ArcTable = annotation_table.Copy()
        def chooser(tupled_arg: tuple[int, CompositeHeader]) -> int | None:
            if tupled_arg[1].is_protocol_column:
                return None

            else: 
                return tupled_arg[0]


        non_protocol_columns: Array[int] = to_array(choose(chooser, indexed(ref_table_1.Headers)))
        ref_table_1.RemoveColumns(non_protocol_columns)
        ref_table_1.RowCount = annotation_table_1.RowCount or 0
        enumerator: Any = get_enumerator(annotation_table_1.Columns)
        try: 
            while enumerator.System_Collections_IEnumerator_MoveNext():
                c: CompositeColumn = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
                ref_table_1.AddColumn(c.Header, c.Cells, None, True)

        finally: 
            dispose(enumerator)

        return ref_table_1

    @staticmethod
    def append(table1: ArcTable, table2: ArcTable) -> ArcTable:
        def get_list(t: ArcTable) -> FSharpList[FSharpList[tuple[CompositeHeader, CompositeCell]]]:
            def _arrow963(__unit: None=None, t: Any=t) -> IEnumerable_1[FSharpList[tuple[CompositeHeader, CompositeCell]]]:
                def _arrow962(row: int) -> FSharpList[tuple[CompositeHeader, CompositeCell]]:
                    def mapping(i: int, c: CompositeCell) -> tuple[CompositeHeader, CompositeCell]:
                        return (t.Headers[i], c)

                    return to_list(map_indexed(mapping, t.GetRow(row, True)))

                return map(_arrow962, range_big_int(0, 1, t.RowCount - 1))

            return to_list(delay(_arrow963))

        pattern_input: tuple[Array[CompositeHeader], ArcTableValues] = Unchecked_alignByHeaders(False, append(get_list(table1), get_list(table2)))
        return ArcTable.from_arc_table_values(table1.Name, pattern_input[0], pattern_input[1])

    def __str__(self, __unit: None=None) -> str:
        this: ArcTable = self
        row_count: int = this.RowCount or 0
        def _arrow972(__unit: None=None) -> IEnumerable_1[str]:
            def _arrow971(__unit: None=None) -> IEnumerable_1[str]:
                def _arrow970(__unit: None=None) -> IEnumerable_1[str]:
                    def _arrow969(__unit: None=None) -> IEnumerable_1[str]:
                        def _arrow964(row_i: int) -> str:
                            return join("\t|\t", map(to_string, this.GetRow(row_i)))

                        def _arrow967(__unit: None=None) -> IEnumerable_1[str]:
                            def _arrow966(__unit: None=None) -> IEnumerable_1[str]:
                                def _arrow965(row_i_1: int) -> str:
                                    return join("\t|\t", map(to_string, this.GetRow(row_i_1)))

                                return map(_arrow965, range_big_int(row_count - 20, 1, row_count - 1))

                            return append_1(singleton("..."), delay(_arrow966))

                        def _arrow968(row_i_2: int) -> str:
                            return join("\t|\t", map(to_string, this.GetRow(row_i_2)))

                        return append_1(map(_arrow964, range_big_int(0, 1, 19)), delay(_arrow967)) if (row_count > 50) else (singleton("No rows") if (row_count == 0) else map(_arrow968, range_big_int(0, 1, row_count - 1)))

                    return append_1(singleton(join("\t|\t", map(to_string, this.Headers))), delay(_arrow969))

                return append_1(singleton("-------------"), delay(_arrow970))

            return append_1(singleton(("Table: " + this.Name) + ""), delay(_arrow971))

        return join("\n", to_list(delay(_arrow972)))

    def StructurallyEquals(self, other: ArcTable) -> bool:
        this: ArcTable = self
        return safe_hash(this) == safe_hash(other)

    def ReferenceEquals(self, other: ArcTable) -> bool:
        this: ArcTable = self
        return this is other

    def __eq__(self, other: Any=None) -> bool:
        this: ArcTable = self
        return this.StructurallyEquals(other) if isinstance(other, ArcTable) else False

    def __hash__(self, __unit: None=None) -> Any:
        this: ArcTable = self
        v_hash: int = safe_hash(this._values) or 0
        return box_hash_array([this.Name, box_hash_seq(this.Headers), v_hash])


ArcTable_reflection = _expr973

def ArcTable__ctor_2D310C9B(name: str, headers: Array[CompositeHeader] | None=None, columns: Array[Array[CompositeCell]] | None=None) -> ArcTable:
    return ArcTable(name, headers, columns)


__all__ = ["ArcTable_reflection"]




from collections.abc import Callable
from typing import Any, TypeVar

from .util import IComparer_1, IEqualityComparer_1, compare, equals, physical_hash, structural_hash


_T_ = TypeVar("_T_")

_T = TypeVar("_T")


def HashIdentity_FromFunctions(hash_1: Callable[[_T], int], eq: Callable[[_T, _T], bool]) -> IEqualityComparer_1[Any]:
    class ObjectExpr0(IEqualityComparer_1[Any]):
        def Equals(self, x: _T_, y: _T_, hash_1: Any = hash_1, eq: Any = eq) -> bool:
            return eq(x, y)

        def GetHashCode(self, x_1: _T_ | None = None, hash_1: Any = hash_1, eq: Any = eq) -> int:
            return hash_1(x_1)

    return ObjectExpr0()


def HashIdentity_Structural(__unit: None = None) -> IEqualityComparer_1[Any]:
    return HashIdentity_FromFunctions(structural_hash, equals)


def HashIdentity_Reference(__unit: None = None) -> IEqualityComparer_1[Any]:
    def _arrow1(e: _T, e_1: _T) -> bool:
        return e == e_1

    return HashIdentity_FromFunctions(physical_hash, _arrow1)


def ComparisonIdentity_FromFunction(comparer: Callable[[_T, _T], int]) -> IComparer_1[Any]:
    class ObjectExpr2(IComparer_1[_T_]):
        def Compare(self, x: _T_, y: _T_, comparer: Any = comparer) -> int:
            return comparer(x, y)

    return ObjectExpr2()


def ComparisonIdentity_Structural(__unit: None = None) -> IComparer_1[Any]:
    return ComparisonIdentity_FromFunction(compare)


__all__ = [
    "HashIdentity_FromFunctions",
    "HashIdentity_Structural",
    "HashIdentity_Reference",
    "ComparisonIdentity_FromFunction",
    "ComparisonIdentity_Structural",
]
